#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/
proc columnize {list {width 80} {nspaces 2}} {
  set maxwidth 0

  set i 0
  foreach item $list {
    set items($i) $item
    set itemwidth [string length $item]
    if {$itemwidth > $maxwidth} {
      set maxwidth $itemwidth
    }
    incr i
  }

  incr maxwidth $nspaces

  set ncol [expr $width/$maxwidth]

  if {$ncol == 0} {
    set ncol 1
  }

  set count [llength $list]
  set nrow [expr $count/$ncol]
  if {[expr $count % $ncol]!=0} {
    incr nrow
  }

  for {set row 0} {$row<$nrow} {incr row} {
    set rows($row) ""
  }

  set maxlen 0
  set itemnum 0
  for {set col 0} {$col<$ncol} {incr col} {
    set newmaxlen $maxlen
    for {set row 0} {$row<$nrow && $itemnum<$count} {incr row} {
      set rows($row) "[format "%-${maxlen}s" $rows($row)]$items($itemnum)"
      incr itemnum
    }
    set maxlen [expr $maxlen + $maxwidth]
  }

  set result {}
  for {set row 0} {$row<$nrow} {incr row} {
    lappend result $rows($row)
  }
  return $result
}

proc getleft {varname count} {
  upvar $varname var
  assert {[info exists var]}
  set left [string range $var 0 [expr {$count-1}]]
  set var [string range $var $count end]
  return $left
}

proc getchar {varname} {
  upvar $varname var
  set char [string index $var 0]
  set var [string range $var 1 end]
  return $char
}

proc string_end {str pattern start} {
  if {[regexp -indices $pattern [string range $str $start end] range]} {
    return [expr {$start+[lindex $range 1]}]
  } else {
    assert{0}
  }
}

proc getstring_index {str delim index} {
  while {1} {
    set char [string index $str $index]
    if {$char==""} {
      return -1
    }
    if {$char=="\\"} {
      incr index
    } elseif {$char==$delim} {
      return $index
    } else {
      set index [string_end $str "\[^\\$delim]*" $index]
    }
    incr index
  }
}

proc getstring {varname delim} {
  upvar $varname var
  assert {[info exists var]}
  set count [getstring_index $var $delim 0]
  set result [getleft var $count]
  getchar var
  return $result
}

proc getgroup {varname delim} {
  upvar $varname var
  set string ""
  while {1} {
    set char [getchar var]
    if {$char=="\\"} {
      set string "$string$char[getchar var]"
    } elseif {$char=="\""} {
      set string "$string$char[getstring var $char]$char"
    } elseif {$char=="'"} {
      set string "$string$char[getstring var $char]$char"
    } elseif {$char==$delim} {
      set string "$string$char"
      break
    } elseif {$char==""} {
      error "Unmatched $delim"
    } else {
      set string "$string$char"
    }
  }
  return $string
}

proc getword {linevar} {
  upvar $linevar line
  set word [lindex $line 0]
  set line [lrange $line 1 end]
  return $word
}

proc string_replace {str start end newstr} {
  set first [string range $str 0 [expr {$start-1}]]
  set last [string range $str [expr {$end+1}] end]
  return "$first$newstr$last"
}

proc process_data {linevar datavar get_line_proc handle_data_proc} {
  upvar $linevar next_line
  upvar $datavar data
  while {1} {
    uplevel $get_line_proc
    set line [string trim $next_line]
    set linelen [string length $line]
    if {$linelen>0} {
      if {[string index $line 0]!="#"} {
	while {1} {
	  if {[string index $line [expr {$linelen-1}]] == "\\"} {
	    set line [string range $line 0 [expr {$linelen-2}]]
	  } elseif {![info complete $line]} {
	    set line "$line\n"
	  } else {
	    break
	  }
	  uplevel $get_line_proc
	  set line "$line$next_line"
	  set linelen [string length $line]
	}
	set data $line
	uplevel $handle_data_proc
      }
    }
  }
}

proc get_next_file_line {filevar linevar} {
  upvar $filevar file
  upvar $linevar line
  gets $file line
  return [expr {![eof $file]}]
}

proc create_line_list {text} {
  set lines [split $text "\n"]
  set fullline ""
  set newlines {}

  foreach line $lines {
    set lineend [string length $line]
    incr lineend -1
    if {[string index $line $lineend] == "\\"} {
      incr lineend -1
      append fullline [string range $line 0 $lineend]
    } else {
      append fullline $line
      if {![info complete $fullline]} {
	append fullline "\n"
      } else {
	lappend newlines $fullline
	set fullline ""
      }
    }
  }
  if {$fullline!=""} {
    error "Incomplete line"
  }
  return $newlines
}

proc read_file {filename} {
  set file [open $filename "r"]
  set text [read -nonewline $file]
  close $file
  return $text
}

proc process_file {filename datavar handler} {
  upvar $datavar data
  if {[catch {
    set text [read_file $filename]
    set lines [create_line_list $text]
    foreach line $lines {
      set data $line
      uplevel $handler
    }
  } message]} {
    puts "Error processing file $filename: $message"
    global errorInfo
    puts "$errorInfo"
  }
}

proc lremove {list element} {
  set index [lsearch -exact $list $element]
  return [lreplace $list $index $index]
}

proc process_lines {lines linevar handler} {
  upvar $linevar item
  process_data line item {
    if {[llength $lines]>0} {
      set line [lindex $lines 0]
      set lines [lreplace $lines 0 0]
    } else {
      break
    }
  } {
    uplevel $handler
  }
}

proc message_dialog {parent title message options default} {
  if {$parent=="."} {
    set dialog .message
  } else {
    set dialog $parent.message
  }
  toplevel $dialog -class Dialog
  wm title $dialog $title
  let w $dialog {
    let w [frame $w.top_frame] {
      let w [label $w.label] {
	$w configure -text $message
	pack $w
      }
      grid $w -row 0 -column 0 -sticky nsew -padx 10 -pady 10
    }
    grid rowconfigure $w 0 -weight 1
    grid columnconfigure $w 0 -weight 1
    let w [frame $w.separator] {
      $w configure -borderwidth 1 -height 2 -relief sunken
      grid $w -row 1 -column 0 -sticky ew
    }
    let w [frame $w.button_frame -borderwidth 2 -relief raised] {
      set button_num 0
      foreach option $options {
	set button $w.button$button_num
	button $button \
	  -text $option \
	  -command [list set message_dialog_button $button_num]
	if {$button_num==$default} {
	  $button configure -default active
	  set default_button $button
	}
	grid $button -row 0 -column $button_num
	grid columnconfigure $w $button_num -weight 1
	incr button_num
      }
      grid $w -row 2 -column 0 -sticky ew
    }
  }
  if {[info exists default_button]} {
    focus $default_button
    bind $dialog <Return> [list set message_dialog_button $default]
  }
  bind $dialog <Destroy> [list set message_dialog_button -1]
  global message_dialog_button
  center_window $dialog $parent
  tkwait variable message_dialog_button
  if {[winfo exists $dialog]} {
    bind $dialog <Destroy> ""
    destroy $dialog
  }
  return $message_dialog_button
}

proc error_dialog {parent message} {
  message_dialog $parent "Error" $message "Ok" 0
}

proc center_windows {window_list over_window} {
  foreach window $window_list {
    wm withdraw $window
  }
  update idletasks
  set max_width 0
  set total_height 0
  foreach window $window_list {
    set window_width [winfo reqwidth $window]
    set window_height [winfo reqheight $window]
    if {$window_width>$max_width} {
      set max_width $window_width
    }
    incr total_height $window_height
  }
  set over_width [winfo width $over_window]
  set over_height [winfo height $over_window]
  set over_x [winfo rootx $over_window]
  set over_y [winfo rooty $over_window]
  set window_x [expr {$over_x+($over_width-$max_width)/2}]
  set window_y [expr {$over_y+($over_height-$total_height)/2}]
  if {$window_x<0} {
    set window_x 0
  }
  if {$window_y<0} {
    set window_y 0
  }
  foreach window $window_list {
    set window_width [winfo reqwidth $window]
    set window_height [winfo reqheight $window]
    set wx [expr {$window_x+($max_width-$window_width)/2}]
    wm geometry $window +$wx+$window_y
    incr window_y $window_height
  }
  foreach window $window_list {
    wm deiconify $window
  }
}

proc center_window {window over_window} {
  center_windows [list $window] $over_window
  if 0 {
    wm withdraw $window
    update idletasks
    set window_width [winfo reqwidth $window]
    set window_height [winfo reqheight $window]
    set over_width [winfo reqwidth $over_window]
    set over_height [winfo reqheight $over_window]
    set over_x [winfo rootx $over_window]
    set over_y [winfo rooty $over_window]
    set window_x [expr {$over_x+($over_width-$window_width)/2}]
    set window_y [expr {$over_y+($over_height-$window_height)/2}]
    wm geometry $window +$window_x+$window_y
    wm deiconify $window
  }
}

proc next_element {list indexvar} {
  upvar $indexvar index
  set value [lindex $list $index]
  incr index
  return $value
}

proc set_frame_sensitivity {w value} {
  if {$value} {
    set state normal
  } else {
    set state disabled
  }
  if {[winfo class $w]=="Label"} {
    if {$state=="normal"} {
      $w configure -foreground black
    } else {
      $w configure -foreground "#a3a3a3"
    }
  } else {
    catch {$w configure -state $state}
  }
  foreach child [winfo children $w] {
    set_frame_sensitivity $child $value
  }
}
