/********************************************************/
/* Gsh Copyright (C) 1999 Vaughn Cato                   */
/* See the file "LICENSE", included with this software, */
/* for license information.                             */
/********************************************************/

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <tcl.h>
#include <tk.h>
#include "spawn.h"
#include "util.h"
#include "termtcl.h"

static char *prog_path = 0;

static int SpawnTermVoidProc(void *client_data,TermOp_t op)
{
  Term *term = (Term *)client_data;

  switch (op) {
    case TERM_NORM_ATTR:
      TermTcl_ResetTags(term);
      break;
    case TERM_TAB:
      TermTcl_Tab(term);
      break;
    case TERM_MOVE_DOWN:
      TermTcl_MoveDown(term);
      break;
    case TERM_CARRIAGE_RETURN:
      TermTcl_CR(term);
      break;
    case TERM_CLEAR_TO_RIGHT:
      TermTcl_ClearToRight(term);
      break;
    case TERM_BACKSPACE:
      TermTcl_Backspace(term);
      break;
    case TERM_HOME_CURSOR:
      TermTcl_HomeCursor(term);
      break;
    case TERM_MOVE_UP:
      TermTcl_MoveUp(term);
      break;
    case TERM_CLEAR_SCREEN:
      TermTcl_ClearScreen(term);
      break;
    case TERM_SAVE_CURSOR:
      TermTcl_SaveCursor(term);
      break;
    case TERM_RESTORE_CURSOR:
      TermTcl_RestoreCursor(term);
      break;
    case TERM_CLEAR_BELOW:
      TermTcl_ClearBelow(term);
      break;
    case TERM_CLEAR_TABS:
      TermTcl_ClearTabs(term);
      break;
    default:
      return 0;
  }
  return 1;
}

static int SpawnTermIntProc(void *client_data,TermOp_t op,int arg1)
{
  Term *term = (Term *)client_data;

  switch (op) {
    case TERM_FGCOLOR_ATTR:
      TermTcl_SetFgColor(term,arg1);
      break;
    case TERM_BGCOLOR_ATTR:
      TermTcl_SetBgColor(term,arg1);
      break;
    case TERM_ALT_CHARSET:
      TermTcl_SetCharset(term,arg1);
      break;
    case TERM_BOLD_ATTR:
      /* Not implemented */
      break;
    case TERM_CURSOR_UP:
      TermTcl_CursorUp(term,arg1);
      break;
    case TERM_CURSOR_DOWN:
      TermTcl_CursorDown(term,arg1);
      break;
    case TERM_CURSOR_LEFT:
      TermTcl_CursorLeft(term,arg1);
      break;
    case TERM_CURSOR_RIGHT:
      TermTcl_CursorRight(term,arg1);
      break;
    case TERM_INVERSE_ATTR:
      TermTcl_SetInverse(term,arg1);
      break;
    case TERM_INSERT_LINE:
      TermTcl_InsertLine(term,arg1);
      break;
    case TERM_INSERT_MODE:
      TermTcl_SetInsertMode(term,arg1);
      break;
    case TERM_SET_TAB:
      TermTcl_SetTab(term,arg1);
      break;
    case TERM_DIM_ATTR:
#if 0
      TermTcl_SetFontTag(term,"dim",arg1);
#endif
      break;
    case TERM_ITALIC_ATTR:
#if 0
      TermTcl_SetFontTag(term,"italic",arg1);
#endif
      break;
    case TERM_UNDERLINE_ATTR:
      TermTcl_SetFontTag(term,"underline",arg1);
      break;
    case TERM_SET_ROW:
      TermTcl_SetRow(term,arg1);
      break;
    case TERM_SET_COL:
      TermTcl_SetCol(term,arg1);
      break;
    case TERM_SET_SCREEN:
      /* Not implemented */
      break;
    case TERM_AUTOWRAP:
      /* Not implemented */
      break;
    case TERM_DELETE_LINE:
      TermTcl_DeleteLine(term,arg1);
      break;
    case TERM_INSERT_CHAR:
      TermTcl_InsertChar(term,arg1);
      break;
    case TERM_DELETE_CHAR:
      TermTcl_DeleteChar(term,arg1);
      break;
    default:
      return 0;
  }
  return 1;
}

static int SpawnTermIntIntProc(void *client_data,TermOp_t op,int arg1,int arg2)
{
  Term *term = (Term *)client_data;

  switch (op) {
    case TERM_SET_CURSOR:
      TermTcl_SetCursor(term,arg1,arg2);
      break;
    case TERM_SET_SCROLL_REGION:
      TermTcl_SetScrollRegion(term,arg1,arg2);
      break;
    case TERM_SET_FONT_STYLE:
    {
      int charset = arg1;
      int term_style = arg2;
      FontStyle_t style;

      switch (term_style) {
	case TERM_DEC_SPECIAL_STYLE:   style = DEC_SPECIAL_STYLE;   break;
	case TERM_UK_STYLE:            style = UK_STYLE;            break;
	case TERM_USASCII_STYLE:       style = USASCII_STYLE;       break;
	case TERM_MULTINATIONAL_STYLE: style = MULTINATIONAL_STYLE; break;
	case TERM_FINNISH_STYLE:       style = FINNISH_STYLE;       break;
	case TERM_GERMAN_STYLE:        style = GERMAN_STYLE;        break;
	default: assert(0);
      }
      TermTcl_SetFontStyle(term,charset,style);
      break;
    }
    default:
      return 0;
  }
  return 1;
}

static void SpawnTermTextProc(void *client_data,const char *text,int len)
{
  Term *term = (Term *)client_data;

  TermTcl_Insert(term,text,len);
}

static void SpawnTermUnexpectedProc(void *client_data,int value)
{
  Term *term = (Term *)client_data;

  TermTcl_Unexpected(term,value);
}


static CommandHandlers spawn_term_cmd_handlers =
  {
    SpawnTermVoidProc,
    SpawnTermIntProc,
    SpawnTermIntIntProc,
    SpawnTermTextProc,
    SpawnTermUnexpectedProc
  };

static int 
  Handle_set_spawn_term(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    SpawnData *spawn_data = 0;
    Term *term = 0;

    if (GetSpawn(interp,argv[1],&spawn_data)!=TCL_OK) {
      return TCL_ERROR;
    }
    term = TermPtr(interp,argv[2]);
    if (!term) {
      return TCL_ERROR;
    }
    SpawnSetDefaultProcs(spawn_data,&spawn_term_cmd_handlers,term);
  }
  return TCL_OK;
}

int Tcl_AppInit(Tcl_Interp *interp)
{
  if (Tcl_Init(interp) == TCL_ERROR)
    return TCL_ERROR;
  Tcl_SetVar(interp,"tcl_interactive","0",TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"restart_command",prog_path,TCL_GLOBAL_ONLY);
  {
    int code = Tk_Init(interp);
    if (code != TCL_OK) {
      fprintf(stderr,"Tk_Init failed\n");
      return code;
    }
  }
  {
    int code = Spawn_Init(interp);
    if (code != TCL_OK) {
      fprintf(stderr,"Spawn_Init failed\n");
      return code;
    }
  }
  {
    int code = TermTcl_Init(interp);
    if (code != TCL_OK) {
      fprintf(stderr,"TermTcl_Init failed\n");
      return code;
    }
    atexit(TermTcl_PrintProfile);
  }
  {
    Tcl_CreateCommand(
      interp,
      "set_spawn_term",
      Handle_set_spawn_term,
      (ClientData)0,
      (Tcl_CmdDeleteProc *)0
    );
  }
  Tcl_StaticPackage(interp,"Tk",Tk_Init,(Tcl_PackageInitProc *)NULL);
  Tcl_StaticPackage(interp,"Spawn",Spawn_Init,(Tcl_PackageInitProc *)NULL);
  return TCL_OK;
}

static char *GetCWD(void)
{
  int size = 1024;

  for (;;) {
    char *result = (char *)CheckAlloc(size);
    if (!getcwd(result,size)) {
      free(result);
      size *= 2;
    }
    else {
      return result;
    }
  }
}

/* Returns an allocated string */
static char *DirName(const char *path)
{
  char *last_slash = strrchr(path,'/');

  if (!last_slash) {
    return strdup(".");
  }
  else {
    int size = last_slash-path;
    char *result = (char *)CheckAlloc(size+1);

    strncpy(result,path,size);
    result[size] = '\0';
    if (strcmp(last_slash+1,".")==0) {
      char *prev_dir = DirName(result);

      free(result);
      result = prev_dir;
    }
    return result;
  }
}

static char *Concat(const char *s1,const char *s2)
{
  int len1 = strlen(s1);
  int len2 = strlen(s2);
  int len = len1+len2;
  char *result = (char *)CheckAlloc(len+1);
  sprintf(result,"%s%s",s1,s2);
  assert(result[len]=='\0');
  return result;
}

static char *FullPath(const char *dir,const char *path);

static char *ExpandLink(const char *path)
{
  char *expanded_path = 0;
  {
    char buffer[1024];
    int len = readlink(path,buffer,(sizeof buffer)-1);
    if (len<0) {
      if (errno==EINVAL) {
	return 0;
      }
      fprintf(stderr,"Trying to expand %s: ",path);
      perror("readlink");
      exit(EXIT_FAILURE);
    }
    buffer[len] = '\0';
    {
      char *dirname = DirName(path);
      expanded_path = FullPath(path,buffer);
      free(dirname);
    }
  }
  if (strcmp(expanded_path,path)==0)  /* To prevent loops */
    return expanded_path;
  {
    char *final_path = ExpandLink(expanded_path);
    if (!final_path) /* Not a link */
      return expanded_path;
    free(expanded_path);
    return final_path;
  }
}

static char *FullPath(const char *dir,const char *path)
{
  if (path[0]=='/')
    return strdup(path);
  {
    char *dir_with_slash = Concat(dir,"/");
    char *result = Concat(dir_with_slash,path);
    free(dir_with_slash);
    return result;
  }
}

static void Append(char **str,const char *appendage)
{
  char *newstr = Concat(*str,appendage);
  free(*str);
  *str = newstr;
}

int main(int argc,char **argv)
{
  char *cwd = GetCWD();

  if (access(argv[0],R_OK|X_OK)==0) {
    prog_path = FullPath(cwd,argv[0]);
  }
  else if (access("/opt/gsh/bin/gsh",R_OK|X_OK)==0) {
    prog_path = FullPath(cwd,"/opt/gsh/bin/gsh");
  }
  else if (access("/usr/local/gsh/bin/gsh",R_OK|X_OK)==0) {
    prog_path = FullPath(cwd,"/usr/local/gsh/bin/gsh");
  }
  else if (access("/usr/local/bin/gsh",R_OK|X_OK)==0) {
    prog_path = FullPath(cwd,"/usr/local/bin/gsh");
  }
  else {
    fprintf(stderr,"Can't find the executable.\n");
    exit(EXIT_FAILURE);
  }

  fprintf(stderr,"prog_path=%s\n",prog_path);
  {
    char *expanded_path = ExpandLink(prog_path);
    if (expanded_path) { /* Was a link */
      free(prog_path);
      prog_path = expanded_path;
    }
  }
  {
    char *prog_dir = DirName(DirName(prog_path));
    {
      static const char *args[4];
      
      Append(&prog_dir,"/lib/gsh.tcl");
      args[0] = argv[0];
      args[1] = "-f";
      args[2] = prog_dir;
      args[3] = NULL;
      {
	int args_count = sizeof args/sizeof *args;
	Tk_Main(args_count,(char **)args,Tcl_AppInit);
      }
    }
    free(prog_dir);
  }
  return 0;
}
