/********************************************************/
/* Gsh Copyright (C) 1999 Vaughn Cato                   */
/* See the file "LICENSE", included with this software, */
/* for license information.                             */
/********************************************************/
#include "spawn.h"

#include <tcl.h>
#include <tk.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#ifdef sun
#  include <stropts.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <assert.h>
#include <termios.h>
#include <sys/ioctl.h>
#include "spawndata.h"
#include "termstate.h"
#include "pty.h"
#include "util.h"

#define DEBUG_TERM 0
  /* If DEBUG_TERM, display each term handler that is called */

static void child_exec(char **argv)
{
  execvp(argv[0],argv);
  perror(argv[0]);
  exit(EXIT_FAILURE);
}

static void (*orig_child_signal_handler)(int) = 0;
static Tcl_AsyncHandler async_handler;

static void set_fl(int fd,int flags)
{
  int val = fcntl(fd,F_GETFL,0);
  
  if (val<0) {
    fprintf(stderr,"fcntl F_GETFL error\n");
    val = 0;
  }
  val |= flags;
  if (fcntl(fd,F_SETFL,val)<0) {
    fprintf(stderr,"fcntl F_SETFL error\n");
  }
}

static int EvalCmdObj(Tcl_Interp *interp,Tcl_Obj *cmd_obj)
{
  int code = Tcl_GlobalEvalObj(interp,cmd_obj);

  if (code!=TCL_OK) {
    fprintf(stderr,"Error executing %s: ",Tcl_GetStringFromObj(cmd_obj,NULL));
    fprintf(stderr,"  %s\n",Tcl_GetStringResult(interp));
    fprintf(stderr,"  %s\n",Tcl_GetVar(interp,"errorInfo",TCL_GLOBAL_ONLY));
  }
  return code;
}

static int ExecIntCmd(Tcl_Interp *interp,const char *cmd,int arg)
{
  Tcl_Obj *obj = Tcl_NewStringObj((char *)cmd,strlen(cmd));
  int code;

#if DEBUG_TERM
  fprintf(stderr,"Executing %s %d:",cmd,arg);
  Tcl_Eval(interp,"update idletasks");
  fgetc(stdin);
#endif

  Tcl_IncrRefCount(obj);
  ListObjAppendInt(obj,arg);
  code = EvalCmdObj(interp,obj);
  Tcl_DecrRefCount(obj);
  return code;
}

static int TermEval(void *data,TermOp_t op)
{
  const char *cmd = 0;
  SpawnData *spawn_data = (SpawnData *)data;
  Tcl_Interp *interp = spawn_data->interp;

  switch (op) {
    case TERM_HOME_CURSOR:
      cmd = spawn_data->term_commands.home_cursor;
      break;
    case TERM_CLEAR_BELOW:
      cmd = spawn_data->term_commands.clear_below;
      break;
    case TERM_CLEAR_ABOVE:
      cmd = spawn_data->term_commands.clear_above;
      break;
    case TERM_CLEAR_TO_RIGHT:
      cmd = spawn_data->term_commands.clear_to_right;
      break;
    case TERM_CLEAR_TO_LEFT:
      cmd = spawn_data->term_commands.clear_to_left;
      break;
    case TERM_CLEAR_SCREEN:
      cmd = spawn_data->term_commands.clear_screen;
      break;
    case TERM_CLEAR_LINE:
      cmd = spawn_data->term_commands.clear_line;
      break;
    case TERM_NORM_ATTR:
      cmd = spawn_data->term_commands.norm_attr;
      break;
    case TERM_SAVE_CURSOR:
      cmd = spawn_data->term_commands.save_cursor;
      break;
    case TERM_RESTORE_CURSOR:
      cmd = spawn_data->term_commands.restore_cursor;
      break;
    case TERM_BELL:
      cmd = spawn_data->term_commands.bell;
      break;
    case TERM_BACKSPACE:
      cmd = spawn_data->term_commands.backspace;
      break;
    case TERM_TAB:
      cmd = spawn_data->term_commands.tab;
      break;
    case TERM_CLEAR_TABS:
      cmd = spawn_data->term_commands.clear_tabs;
      break;
    case TERM_CARRIAGE_RETURN:
      cmd = spawn_data->term_commands.carriage_return;
      break;
    case TERM_MOVE_UP:
      cmd = spawn_data->term_commands.move_up;
      break;
    case TERM_MOVE_DOWN:
      cmd = spawn_data->term_commands.move_down;
      break;
    case TERM_EOF:
      cmd = spawn_data->term_commands.eof;
      break;
    default:
      assert(0);
  }
  if (cmd) {
    int code;
    Tcl_Obj *cmd_obj = Tcl_NewStringObj((char *)cmd,strlen(cmd));

#if DEBUG_TERM
    fprintf(stderr,"Executing %s:",cmd);
    Tcl_Eval(interp,"update idletasks");
    fgetc(stdin);
#endif

    Tcl_IncrRefCount(cmd_obj);
    code = EvalCmdObj(interp,cmd_obj);
    Tcl_DecrRefCount(cmd_obj);
    return (code==TCL_OK);
  }
  else {
    int (*void_proc)(void *,TermOp_t) =
      spawn_data->default_procs.void_proc;

    if (void_proc) {
      return (*void_proc)(spawn_data->default_data,op);
    }
  }
  return 0;
}

static int TermEvalInt(void *data,TermOp_t op,int arg1)
{
  const char *cmd = 0;
  SpawnData *spawn_data = (SpawnData *)data;
  Tcl_Interp *interp = spawn_data->interp;

  switch (op) {
    case TERM_APP_CURSOR_KEYS:
      cmd = spawn_data->term_commands.app_cursor_keys;
      break;
    case TERM_AUTOWRAP:
      cmd = spawn_data->term_commands.autowrap;
      break;
    case TERM_SET_SCREEN:
      cmd = spawn_data->term_commands.set_screen;
      break;
    case TERM_CURSOR_VISIBLE:
      cmd = spawn_data->term_commands.cursor_visible;
      break;
    case TERM_MOUSE_REPORTING:
      cmd = spawn_data->term_commands.mouse_reporting;
      break;
    case TERM_MOUSE_HIGHLIGHTING:
      cmd = spawn_data->term_commands.mouse_highlighting;
      break;
    case TERM_SET_ROW:
      cmd = spawn_data->term_commands.set_row;
      break;
    case TERM_SET_COL:
      cmd = spawn_data->term_commands.set_col;
      break;
    case TERM_SET_TAB:
      cmd = spawn_data->term_commands.set_tab;
      break;
    case TERM_INSERT_LINE:
      cmd = spawn_data->term_commands.insert_line;
      break;
    case TERM_DELETE_LINE:
      cmd = spawn_data->term_commands.delete_line;
      break;
    case TERM_INSERT_CHAR:
      cmd = spawn_data->term_commands.insert_char;
      break;
    case TERM_DELETE_CHAR:
      cmd = spawn_data->term_commands.delete_char;
      break;
    case TERM_CURSOR_UP:
      cmd = spawn_data->term_commands.cursor_up;
      break;
    case TERM_CURSOR_DOWN:
      cmd = spawn_data->term_commands.cursor_down;
      break;
    case TERM_CURSOR_RIGHT:
      cmd = spawn_data->term_commands.cursor_right;
      break;
    case TERM_CURSOR_LEFT:
      cmd = spawn_data->term_commands.cursor_left;
      break;
    case TERM_APP_KEYPAD:
      cmd = spawn_data->term_commands.app_keypad;
      break;
    case TERM_ALT_CHARSET:
      cmd = spawn_data->term_commands.set_charset;
      break;
    case TERM_BOLD_ATTR:
      cmd = spawn_data->term_commands.bold_attr;
      break;
    case TERM_DIM_ATTR:
      cmd = spawn_data->term_commands.dim_attr;
      break;
    case TERM_ITALIC_ATTR:
      cmd = spawn_data->term_commands.italic_attr;
      break;
    case TERM_UNDERLINE_ATTR:
      cmd = spawn_data->term_commands.underline_attr;
      break;
    case TERM_BLINK_ATTR:
      cmd = spawn_data->term_commands.blink_attr;
      break;
    case TERM_INVERSE_ATTR:
      cmd = spawn_data->term_commands.inverse_attr;
      break;
    case TERM_FGCOLOR_ATTR:
      cmd = spawn_data->term_commands.fgcolor_attr;
      break;
    case TERM_BGCOLOR_ATTR:
      cmd = spawn_data->term_commands.bgcolor_attr;
      break;
    case TERM_INSERT_MODE:
      cmd = spawn_data->term_commands.insert_mode;
      break;
    default:
      assert(0);
  }
  if (cmd) {
    int code = ExecIntCmd(interp,cmd,arg1);
    return (code==TCL_OK);
  }
  else {
    int (*int_proc)(void *,TermOp_t,int) = spawn_data->default_procs.int_proc;

    if (int_proc) {
      return (*int_proc)(spawn_data->default_data,op,arg1);
    }
  }
  return 0;
}

static void TermUnexpected(void *data,int c)
{
  SpawnData *spawn_data = (SpawnData *)data;
  const char *cmd = spawn_data->term_commands.unexpected;

  assert(c>=0);
  if (cmd) {
    ExecIntCmd(spawn_data->interp,cmd,c);
  }
  else {
    void (*unexpected_proc)(void *,int) =
      spawn_data->default_procs.unexpected_proc;

    if (unexpected_proc) {
      return (*unexpected_proc)(spawn_data->default_data,c);
    }
  }
}

static int TermEval2Int(void *data,TermOp_t op,int arg1,int arg2)
{
  SpawnData *spawn_data = (SpawnData *)data;
  Tcl_Interp *interp = spawn_data->interp;
  const char *cmd = 0;

  switch (op) {
    case TERM_SET_CURSOR:
      cmd = spawn_data->term_commands.set_cursor;
      break;
    case TERM_SET_SCROLL_REGION:
      cmd = spawn_data->term_commands.set_scroll_region;
      break;
    case TERM_SET_FONT_STYLE:
    {
      int charset = arg1;
      int style = arg2;
      const char *style_name = 0;

      cmd = spawn_data->term_commands.set_font_style;
      if (cmd) {
	switch (style) {
	  case TERM_DEC_SPECIAL_STYLE:   style_name="dec_special";   break;
	  case TERM_UK_STYLE:            style_name="uk";            break;
	  case TERM_USASCII_STYLE:       style_name="usascii";       break;
	  case TERM_MULTINATIONAL_STYLE: style_name="multinational"; break;
	  case TERM_FINNISH_STYLE:       style_name="finnish";       break;
	  case TERM_GERMAN_STYLE:        style_name="german";        break;
	  default: assert(0);
	}
	{
	  int code;
	  Tcl_Obj *cmd_obj = Tcl_NewStringObj((char *)cmd,strlen(cmd));

	  Tcl_IncrRefCount(cmd_obj);
	  ListObjAppendInt(cmd_obj,charset);
	  ListObjAppendString(cmd_obj,style_name);
	  code = EvalCmdObj(interp,cmd_obj);
	  Tcl_DecrRefCount(cmd_obj);
	  return (code==TCL_OK);
	}
      }
      break;
    }
    default:
      assert(0);
  }

  if (cmd) {
    int code;
    Tcl_Obj *cmd_obj = Tcl_NewStringObj((char *)cmd,strlen(cmd));

#if DEBUG_TERM
    fprintf(stderr,"Executing %s %d %d:",cmd,arg1,arg2);
    Tcl_Eval(interp,"update idletasks");
    fgetc(stdin);
#endif

    Tcl_IncrRefCount(cmd_obj);
    ListObjAppendInt(cmd_obj,arg1);
    ListObjAppendInt(cmd_obj,arg2);
    code = EvalCmdObj(interp,cmd_obj);
    Tcl_DecrRefCount(cmd_obj);
    return code==TCL_OK;
  }
  else {
    int (*int_int_proc)(void *,TermOp_t,int,int) =
      spawn_data->default_procs.int_int_proc;

    if (int_int_proc) {
      return (*int_int_proc)(spawn_data->default_data,op,arg1,arg2);
    }
  }
  return 0;
}

static void TermText(void *data,const char *text,int len)
{
  const char *cmd = 0;
  SpawnData *spawn_data = (SpawnData *)data;
  Tcl_Interp *interp = spawn_data->interp;

  cmd = spawn_data->term_commands.text;
  if (cmd) {
    Tcl_Obj *cmd_obj = Tcl_NewStringObj((char *)cmd,strlen(cmd));

#if DEBUG_TERM
    fprintf(stderr,"Executing %s {%.*s}:",cmd,len,text);
    Tcl_Eval(interp,"update idletasks");
    fgetc(stdin);
#endif
    Tcl_IncrRefCount(cmd_obj);
    ListObjAppendBytes(cmd_obj,text,len);
    EvalCmdObj(interp,cmd_obj);
    Tcl_DecrRefCount(cmd_obj);
  }
  else {
    void (*text_proc)(void *,const char *,int) =
      spawn_data->default_procs.text_proc;

    if (text_proc) {
      (*text_proc)(spawn_data->default_data,text,len);
    }
    else {
      assert(0);
    }
  }
}

static CommandHandlers 
  spawn_cmd_handlers = 
    { TermEval,TermEvalInt,TermEval2Int,TermText,TermUnexpected };

static void ReadMaster(ClientData client_data,int mask)
{
  SpawnData *spawn_data = (SpawnData *)client_data;

  assert(!spawn_data->reading_master);
    /* This shouldn't be possible since we are removing the file handler */
    /* while the input characters are being processed */

  spawn_data->reading_master = 1;
  assert(spawn_data->interp);
  assert(spawn_data->master_fd!=-1);

  if (mask & TCL_READABLE) {
    unsigned char buffer[1024];
    int bufsize = sizeof buffer;
    int readsize = read(spawn_data->master_fd,buffer,bufsize);

    if (readsize<1) {
      Tcl_DeleteFileHandler(spawn_data->master_fd);
      TermState_EOF(&spawn_data->term_state);
      assert(spawn_data->master_fd>=0);
      close(spawn_data->master_fd);
      spawn_data->master_fd = -1;
    }
    else {
      Tcl_DeleteFileHandler(spawn_data->master_fd);
	/* Temporarily remove the file handler so we won't have new */
	/* characters coming in while we are processing the ones we just */
	/* read. */

      while (readsize<bufsize) {
	{
	  struct timeval timeout = {0,1};
	  fd_set rdfds;
	  FD_ZERO(&rdfds);
	  FD_SET(spawn_data->master_fd,&rdfds);
	  select(spawn_data->master_fd+1,&rdfds,NULL,NULL,&timeout);
	  if (!FD_ISSET(spawn_data->master_fd,&rdfds)) {
	    break;
	  }
	}
	{
	  int count = 
	    read(spawn_data->master_fd,buffer+readsize,bufsize-readsize);
	  if (count<=0)
	    break;
	  readsize += count;
        }
      }
      {
	int i = 0;
	for (;i<readsize;++i) {
	  TermState_AddChar(&spawn_data->term_state,buffer[i]);
	}
      }

      TermState_Flush(&spawn_data->term_state);

      Tcl_CreateFileHandler(
	spawn_data->master_fd,TCL_READABLE|TCL_EXCEPTION,ReadMaster,spawn_data
      );
    }
  }
  else {
    fprintf(stderr,"ReadMaster called with nothing ready\n");
    assert(0);
  }
  assert(spawn_data->reading_master);
  spawn_data->reading_master = 0;

  if (spawn_data->master_fd == -1) {
    DeleteSpawnData(spawn_data);
  }
}

static void MakeRaw(struct termios *tios)
{
  tios->c_lflag &= ~ICANON;
  tios->c_cc[VMIN] = 1;
  tios->c_cc[VTIME] = 0;
}

static int ReadSlave(SpawnData *spawn_data,char *buffer,int bufsize)
{
  struct termios normattr;
  struct termios termattr;
  int readsize;

  tcgetattr(spawn_data->slave_fd,&normattr);
  termattr = normattr;
  MakeRaw(&termattr);
  termattr.c_cc[VEOF] = 0;
  tcsetattr(spawn_data->slave_fd,TCSANOW,&termattr);
  set_fl(spawn_data->slave_fd,O_NONBLOCK);
  readsize = read(spawn_data->slave_fd,buffer,bufsize-1);
  tcsetattr(spawn_data->slave_fd,TCSANOW,&normattr);
  if (readsize<0) {
    readsize = 0;
  }

  /* eof characters get turned into nulls.  Change them back. */
  /* I'm not sure if this is normal, or if it just happens under linux. */
  { 
    int i = 0;
    for (;i<readsize;++i) {
      if (buffer[i]==0) {
	buffer[i] = normattr.c_cc[VEOF];
      }
    }
  }
  buffer[readsize] = '\0';
  return readsize;
}

static const struct {
  const char *name;
  int signum;
} signals[] = {
  {"HUP",   SIGHUP},
  {"INT",   SIGINT},
  {"QUIT",  SIGQUIT},
  {"ILL",   SIGILL},
  {"TRAP",  SIGTRAP},
  {"ABRT",  SIGABRT},
  {"BUS",   SIGBUS},
  {"FPE",   SIGFPE},
  {"KILL",  SIGKILL},
  {"USR1",  SIGUSR1},
  {"SEGV",  SIGSEGV},
  {"USR2",  SIGUSR2},
  {"PIPE",  SIGPIPE},
  {"ALRM",  SIGALRM},
  {"TERM",  SIGTERM},
  {"CHLD",  SIGCHLD},
  {"CONT",  SIGCONT},
  {"STOP",  SIGSTOP},
  {"TSTP",  SIGTSTP},
  {"TTIN",  SIGTTIN},
  {"TTOU",  SIGTTOU},
  {"URG",   SIGURG},
  {"XCPU",  SIGXCPU},
  {"XFSZ",  SIGXFSZ},
  {"VTALRM",SIGVTALRM},
  {"PROF",  SIGPROF},
  {"WINCH", SIGWINCH},
  {"IO",    SIGIO},
  {"PWR",   SIGPWR},
  {NULL,-1}
};

static const char *SignalName(int signum)
{
  int i = 0;

  for (;signals[i].name;++i) {
    if (signals[i].signum==signum) {
      return signals[i].name;
    }
  }
  return 0;
}


static void SpawnExit(SpawnData *spawn_data,int exit_signal,int exit_code)
{
  Tcl_Interp *interp = spawn_data->interp;

  assert(spawn_data->slave_fd!=-1);
  assert(spawn_data->child_running);
  spawn_data->child_running = 0;
  spawn_data->child_pid = -1;
  spawn_data->child_exit_handled = 1;
  {
    assert(!spawn_data->child_running);
    {
      const char *cmdname = "spawn_exit";
      int code;
      {
	Tcl_Obj *cmd_obj = Tcl_NewObj();
	Tcl_IncrRefCount(cmd_obj);
	Tcl_AppendStringsToObj(cmd_obj,cmdname,NULL);
	Tcl_ListObjAppendElement(
	  NULL,cmd_obj,Tcl_NewIntObj(SpawnId(spawn_data))
	);
	if (exit_signal!=0) {
	  ListObjAppendString(cmd_obj,SignalName(exit_signal));
	} else {
	  ListObjAppendInt(cmd_obj,exit_code);
	}
	code = Tcl_GlobalEvalObj(interp,cmd_obj);
	Tcl_DecrRefCount(cmd_obj);
      }
      if (code!=TCL_OK) {
	fprintf(stderr,"Error executing %s: ",cmdname);
	fprintf(stderr,"%s\n",Tcl_GetStringResult(interp));
	fprintf(
	  stderr,"  %s\n",Tcl_GetVar(interp,"errorInfo",TCL_GLOBAL_ONLY)
	);
      }
    }
  }
}

static void SIGCHLDHandler(int signum);

static void CheckChildExit(ClientData client_data)
{
  SpawnData *spawn_data = 0;
  int exit_code = 0;
  int exit_signal = 0;
  {
    int status;
    pid_t exited_pid = waitpid(-1,&status,WNOHANG);

    if (exited_pid==-1) {
      /* Errors handled below */
    }
    else if (WIFEXITED(status)) {
      fprintf(stderr,"Process %d exited\n",(int)exited_pid);
      exit_code=WEXITSTATUS(status);
    } 
    else if (WIFSTOPPED(status)) {
      fprintf(stderr,"Process %d stopped\n",(int)exited_pid);
    }
    else if (WIFSIGNALED(status)) {
      exit_signal = WTERMSIG(status);
      fprintf(
        stderr,
	"Process %d was signalled with %d\n",(int)exited_pid,exit_signal
      );
    }
    else {
      fprintf(stderr,"Process %d exited abnormally\n",(int)exited_pid);
    }

    if (exited_pid>0) {
      spawn_data = FirstSpawnData();
      while (spawn_data) {
	if (spawn_data->child_pid==exited_pid) {
	  break;
	}
	spawn_data = spawn_data->next;
      }
      if (!spawn_data) {
	fprintf(stderr,"Process %d is not in spawn list\n",(int)exited_pid);
      }
    }
    else if (exited_pid==0) {
      fprintf(stderr,"No child processes\n");
    }
    else if (exited_pid==-1 && errno!=EINTR) {
      perror("waitpid");
    }
    else {
      fprintf(stderr,"No child processes\n");
    }
  }
  if (spawn_data) {
    SpawnExit(spawn_data,exit_signal,exit_code);
  }
#ifdef sgi
  signal(SIGCHLD,SIGCHLDHandler);
#endif
}

static int AsyncHandler(ClientData client_data,Tcl_Interp *interp,int code)
{
  Tcl_DoWhenIdle(CheckChildExit,0);
  return code;
}

static void SIGCHLDHandler(int signum)
{
  Tcl_AsyncMark(async_handler);
  if (orig_child_signal_handler) {
    (*orig_child_signal_handler)(signum);
  }
}

static void OpenPty(SpawnData *spawn_data)
{
  assert(spawn_data->master_fd==-1);
  assert(spawn_data->slave_fd==-1);

#ifdef sun
  {
    spawn_data->master_fd = open("/dev/ptmx", O_RDWR);
    if (spawn_data->master_fd == -1) {
      fprintf(stderr,"Error opening pty\n");
      perror("open: /dev/ptmx");
      exit(EXIT_FAILURE);
    }
    grantpt(spawn_data->master_fd);
    unlockpt(spawn_data->master_fd);
    spawn_data->slave_fd = open(ptsname(spawn_data->master_fd), O_RDWR); 
    if (spawn_data->slave_fd==-1) {
      fprintf(stderr,"Error opening slave\n");
      fprintf(stderr, "open: ");
      perror(ptsname(spawn_data->master_fd));
      exit(EXIT_FAILURE);
    }
    /* The following are "to get terminal semantics" */
    ioctl(spawn_data->slave_fd, I_PUSH, "ptem");
    ioctl(spawn_data->slave_fd, I_PUSH, "ldterm");
  }
#endif /* sun */
#ifdef sgi
  {
    char *slave_name =
      _getpty(&spawn_data->master_fd,O_RDWR,S_IREAD|S_IWRITE,0);
    if (!slave_name) {
      fprintf(stderr,"Error opening pty\n");
      perror("_getpty");
      exit(EXIT_FAILURE);
    }
    spawn_data->slave_fd = open(slave_name,O_RDWR);
    if (spawn_data->slave_fd==-1) {
      fprintf(stderr,"Error opening slave\n");
      exit(EXIT_FAILURE);
    }
  }
#endif
#if !defined(sun) && !defined(sgi)
  if (open_pty_pair(&spawn_data->master_fd,&spawn_data->slave_fd)!=0) {
    exit(EXIT_FAILURE);
  }
#endif
  set_fl(spawn_data->master_fd,O_NONBLOCK);
  {
    struct termios termattr;

    tcgetattr(spawn_data->slave_fd,&termattr);
    termattr.c_cc[VINTR] = 3;
    termattr.c_cc[VERASE] = 8;
    tcsetattr(spawn_data->slave_fd,TCSANOW,&termattr);
  }

  /* If the close-on-exec flag is not set for these file descriptors, they */
  /* will remain open in the child processes.  If two child processes are */
  /* started, and the first one finishes.  The master won't get EOF until */
  /* the second one finishes because the second one will have the slave_fd */
  /* open also. */
  fcntl(spawn_data->master_fd,F_SETFD,1);
  fcntl(spawn_data->slave_fd,F_SETFD,1);

  strcpy(spawn_data->slave_name,ttyname(spawn_data->slave_fd));
}

static void SetObjVar(Tcl_Interp *interp,char *name,Tcl_Obj *value_obj)
{
  Tcl_Obj *name_obj = Tcl_NewObj();
  Tcl_IncrRefCount(name_obj);
  Tcl_AppendStringsToObj(name_obj,name,NULL);
  Tcl_ObjSetVar2(interp,name_obj,NULL,value_obj,0);
  Tcl_DecrRefCount(name_obj);
}

static void PrependResult(Tcl_Interp *interp,const char *str)
{
  char *result_str = Tcl_GetStringResult(interp);
  int new_result_size = strlen(str)+strlen(result_str)+1;
  char *new_result = Tcl_Alloc(new_result_size);

  sprintf(new_result,"%s%s",str,result_str);
  assert(new_result[new_result_size-1]=='\0');
  Tcl_SetResult(interp,new_result,TCL_DYNAMIC);
}

int GetSpawn(Tcl_Interp *interp,char *str,SpawnData **result_ptr)
{
  int spawn_id;
  int code = Tcl_GetInt(interp,str,&spawn_id);

  if (code==TCL_OK) {
    *result_ptr = SpawnWithId(spawn_id);
    if (!*result_ptr) {
      Tcl_SetResult(interp,"No spawn has id ",TCL_STATIC);
      Tcl_AppendResult(interp,str,NULL);
      code = TCL_ERROR;
    }
  }

  if (code!=TCL_OK) {
    PrependResult(interp,"Invalid spawn id specified for -i: ");
  }

  return code;
}

static int 
  GetNextArg(
    Tcl_Interp *interp,int argc,char **argv,int *argnum,const char *name
  )
{
  ++*argnum;
  if (*argnum>=argc) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,"No value specified for ",name,NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

static int 
  ParseOptions(
    Tcl_Interp *interp,
    int argc,char **argv,int *argnum_ptr,SpawnData **spawn_data_ptr
  )
{
  int argnum = *argnum_ptr;
  SpawnData *spawn_data = FirstSpawnData();

  if (!spawn_data) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,"No active spawn",NULL);
    return TCL_ERROR;
  }

  while (argnum<argc) {
    if (strcmp(argv[argnum],"-i")==0) {
      {
	int code = GetNextArg(interp,argc,argv,&argnum,"-i option");
	if (code!=TCL_OK) {
	  return code;
	}
      }
      {
	int code = GetSpawn(interp,argv[argnum],&spawn_data);
	
	if (code!=TCL_OK) {
	  return code;
	}
	++argnum;
      }
    }
    else if (strcmp(argv[argnum],"--")==0) {
      ++argnum;
      break;
    }
    else {
      break;
    }
  }
  *argnum_ptr = argnum;
  *spawn_data_ptr = spawn_data;
  return TCL_OK;
}

static int child_signal = -1;

static void ChildSignalHandler(int signum)
{
  child_signal = signum;
}

static int caught_sigint;

static void HandleSIGINT(int signum) 
{
  caught_sigint = 1;
}

static int 
  Handle_spawn_exec(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  int argnum = 1;
  SpawnData *spawn_data = 0;
  
  {
    int code = ParseOptions(interp,argc,argv,&argnum,&spawn_data);
    if (code!=TCL_OK)
      return code;
  }
  {
    /* Attempt to fix problem with Tcl8.1 */
    void (*old_signal_handler)(int) = 
      signal(SIGCHLD,SIGCHLDHandler);
    if (old_signal_handler!=SIGCHLDHandler) {
      fprintf(stderr,"SIGCHLDHandler was removed, reinstalling\n");
      if (old_signal_handler==SIG_ERR) {
	fprintf(stderr,"Error installing signal handler\n");
	perror("signal");
      }
      orig_child_signal_handler = old_signal_handler;
    }
  }
  {
    int child_pid = fork();
    if (child_pid==0) {
      /* child process */

      /* Set signals back to their default behavior */
      /* signal(SIGINT,SIG_DFL); */
      signal(SIGINT,HandleSIGINT);
      signal(SIGQUIT,SIG_DFL);
      signal(SIGCHLD,SIG_DFL);

      /* Close all open files so the child won't inherit them.  An example */
      /* problem is the X11 display socket.  If the child inheretid this, */
      /* then the display would be left open if a child was running even if */
      /* gsh exited. */
      {
	int fdlimit = sysconf(_SC_OPEN_MAX);
	int i = 0;

	for (;i<fdlimit;++i) {
	  close(i);
	}
      }
      setsid();
      {
	/* Open slave here after setsid so that it becomes the controlling */
	/* terminal for this process */
	int slave = open(spawn_data->slave_name,O_RDWR);
	if (slave<0) {
	  perror("open slave");
	  exit(EXIT_FAILURE);
	}
	ioctl(slave,TIOCSWINSZ,&spawn_data->window_size);
	dup(slave);
	dup(slave);
	dup(slave);
      }
      /* Turn off terminal i/o stop signals so that the tcsetpgrp will work */
      {
	signal(SIGTTIN,SIG_IGN);
	signal(SIGTTOU,SIG_IGN);
      }
      /* Fork the grandchild that will be the process group leader of a new */
      /* foreground process group */
      {
	int grandchild_pid = fork();

	/* This will catch continue signals so they can be redirected to */
	/* the grandchild.  The variable child_signal will be set. */
	if (grandchild_pid==0) {
	  int pid = getpid();
	  /* In the grandchild process */
	  /* Make a new process group for the grandchild */
	  {
	    int result = setpgid(0,pid);
	    if (result!=0) {
	      fprintf(stderr,"grandchild setpgid: result=%d\n",result);
	    }
	  }
	  /* Make the grandchild process group be the foreground process */
	  /* group */
	  {
	    int result = tcsetpgrp(0,pid);
	    if (result!=0) {
	      fprintf(stderr,"grandchild tcsetpgrp: result = %d\n",result);
	    }
	  }
	  /* Turn terminal i/o stop signals back on */
	  {
	    signal(SIGTTIN,SIG_DFL);
	    signal(SIGTTOU,SIG_DFL);
	  }
	  child_exec(argv+argnum); /* doesn't return */
	}
	child_signal = -1;
	{
	  struct sigaction act;

	  act.sa_handler=ChildSignalHandler;
	  sigemptyset(&act.sa_mask);
	  act.sa_flags=0;
	  sigaction(SIGCONT,&act,0);
	}
	{
	  int exit_code = 0;
	  int status;

	  for (;;) {
	    pid_t pid;

	    pid = waitpid(grandchild_pid,&status,WUNTRACED);
	    
	    if (pid!=grandchild_pid) {
	      if (child_signal==SIGCONT) {
		kill(grandchild_pid,SIGCONT);
	      }
	      else {
		fprintf(stderr,"waitpid returned %d\n",(int)pid);
		break;
	      }
	    }
	    else {
	      if (WIFSIGNALED(status)) {
#if 1
		fprintf(
		  stderr,
		  "process killed with signal %d\n",(int)WTERMSIG(status)
		);
#endif
		break;
	      } 
	      else if (WIFEXITED(status)) {
		exit_code = WEXITSTATUS(status);
#if 0
		fprintf(stderr,"process exited with status %d\n",exit_code);
#endif
		break;
	      }
	      else {
		if (WSTOPSIG(status)) {
		  fprintf(stderr,"(stopped)\n");
		}
		else {
#if 0
		  fprintf(stderr,"process exited abnormally\n");
#endif
		  break;
		}
	      }
	    }
	  }
	  /* Make the session leader be the foreground process group again. */
	  /* This makes anything the child forked be in the background */
	  /* process group, so it won't get SIGHUP when the session leader */
	  /* dies. */
	  {
	    int pgrp = getpgrp();
	    int result = tcsetpgrp(0,pgrp);
	    if (result!=0) {
	      perror("tcsetpgrp");
	    }
	  }
	  signal(SIGINT,SIG_DFL);
	  if (caught_sigint ||
	       (WIFSIGNALED(status) && WTERMSIG(status)==SIGINT)) {
	    /* Signal self so the parent can tell that the child was  */
	    /* signalled */
	    kill(getpid(),SIGINT);
	  }
	  /* Use _exit to prevent atexit cleanup operations */
	  _exit(exit_code);
	}
      }
    }
    else {
      /* parent */
      spawn_data->child_running = 1;
      spawn_data->child_exit_handled = 0;
      spawn_data->child_pid = child_pid;
      fprintf(stderr,"Process %d started\n",child_pid);
    }
  }
  Tcl_SetObjResult(interp,Tcl_NewIntObj(spawn_data->child_pid));
  return TCL_OK;
}

static int
  Handle_spawn_delete(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  int argnum = 1;
  SpawnData *spawn_data = 0;

  {
    int code = ParseOptions(interp,argc,argv,&argnum,&spawn_data);
    if (code!=TCL_OK) {
      return code;
    }
  }

  if (argnum!=argc) {
    Tcl_SetResult(interp,"Usage: spawn_delete [-i <spawn id>]",TCL_STATIC);
    return TCL_ERROR;
  }

  if (spawn_data->master_fd!=-1) {
    Tcl_DeleteFileHandler(spawn_data->master_fd);
    close(spawn_data->master_fd);
    spawn_data->master_fd = -1;
  }
  if (spawn_data->slave_fd!=-1) {
    close(spawn_data->slave_fd);
    spawn_data->slave_fd = -1;
  }
  spawn_data->child_pid = -1;
  spawn_data->child_running = 0;
  spawn_data->child_exit_handled = 1;
  DeleteSpawnData(spawn_data);
  return TCL_OK;
}

static int 
  Handle_spawn_close(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  int argnum = 1;
  SpawnData *spawn_data = 0;

  {
    int code = ParseOptions(interp,argc,argv,&argnum,&spawn_data);
    if (code!=TCL_OK) {
      return code;
    }
  }

  if (argnum!=argc) {
    Tcl_SetResult(interp,"Usage: spawn_close [-i <spawn id>]",TCL_STATIC);
    return TCL_ERROR;
  }
#if 0
  if (spawn_data->child_running) {
    Tcl_SetResult(interp,"Cannot close pty on a running spawn",TCL_STATIC);
    return TCL_ERROR;
  }
#endif
  {
    char buffer[1024];
    ReadSlave(spawn_data,buffer,sizeof buffer);

    assert(spawn_data->slave_fd!=-1);
    close(spawn_data->slave_fd);
    spawn_data->slave_fd = -1;
    Tcl_SetResult(interp,buffer,TCL_VOLATILE);
  }
  return TCL_OK;
}

static int 
  Handle_spawn_send(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  int error = 0;
  SpawnData *spawn_data = 0;
  int argnum = 1;
  char *key = 0;

  {
    int code = ParseOptions(interp,argc,argv,&argnum,&spawn_data);
    
    if (code!=TCL_OK) {
      return code;
    }
  }

  if (argnum>=argc) {
    error = 1;
  } 
  else {
    key = argv[argnum];
    ++argnum;
  }
  if (argnum<argc) {
    error = 1;
  }
  if (error) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,"Usage: ",argv[0]," [-i <spawn_id>] <key>",NULL);
    return TCL_ERROR;
  }
  write(spawn_data->master_fd,key,strlen(key));
  return TCL_OK;
}

static void RemoveTopLevelNewlines(char *str)
{
  char *p = str;
  int brace_count = 0;

  for (;*p!='\0';++p) {
    if (*p=='{')
      ++brace_count;
    else if (*p=='}' && brace_count>0)
      --brace_count;
    else if (*p=='"') {
      ++p;
      while (*p!='\0' && *p!='"') {
	if (*p=='\\' && *(p+1)!='\0') {
	  ++p;
	}
	++p;
      }
      if (*p=='\0')
	break;
    }
    else if (*p=='\n' && brace_count==0)
      *p = ' ';
  }
}

static int
  Handle_spawn_set_handlers(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  SpawnData *spawn_data = FirstSpawnData();
  int not_braced = 0;
  int argnum = 1;

  if (!spawn_data) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,"No active spawn",NULL);
    return TCL_ERROR;
  }

  while (argnum<argc) {
    if (argv[argnum][0]!='-')
      break;
    if (strcmp(argv[argnum],"--")==0) {
      ++argnum;
      break;
    }
    if (strcmp(argv[argnum],"-i")==0) {
      ++argnum;
      {
	int code = GetSpawn(interp,argv[argnum],&spawn_data);
	if (code!=TCL_OK) {
	  return code;
	}
      }
      ++argnum;
    }
    else if (strcmp(argv[argnum],"-nobrace")==0) {
      not_braced = 1;
      ++argnum;
    }
  }
  if (!not_braced) {
    if (argnum!=argc-1) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(
	interp,"Usage: ",argv[0]," -i <spawn_id> <term command list>"
      );
      return TCL_ERROR;
    }
    /* re-execute the command with a -nobrace option and the list expanded */
    /* into seperate parameters */
    {
      int code;
      {
	char *newcmd = 0;
	{
	  char **new_argv = (char **)CheckAlloc((argc+1)*sizeof(char *));
	  int new_argc = argc+1;
	  {
	    int i = 0;
	    for (;i<argc-1;++i) {
	      new_argv[i] = argv[i];
	    }
	    new_argv[i] = "-nobrace";
	    new_argv[i+1] = argv[argc-1];
	    assert(i+2==new_argc);
	  }
	  newcmd = Tcl_Concat(new_argc,new_argv);
	  free(new_argv);
	}

	RemoveTopLevelNewlines(newcmd);

	code = Tcl_Eval(interp,newcmd);
	Tcl_Free(newcmd);
      }
      return code;
    }
  }
  {
    for (;argnum+1<argc;argnum+=2) {
      const char *cmdname = argv[argnum];
      const char *cmdbody = argv[argnum+1];
      {
	int j = 0;
	for (;term_command_members[j].name;++j) {
	  if (strcmp(cmdname,term_command_members[j].name)==0) {
	    int offset = term_command_members[j].offset;
	    char **cmdptr = 
	      (char **)((char *)&spawn_data->term_commands+offset);
	    if (*cmdptr) {
	      free(*cmdptr);
	    }
	    *cmdptr = strdup(cmdbody);
	    break;
	  }
	}
	if (!term_command_members[j].name) {
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp,"Invalid term operation: ",cmdname,NULL);
	  return TCL_ERROR;
	}
      }
    }
    if (argnum<argc) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"Extra command specified: ",argv[argnum],NULL);
      return TCL_ERROR;
    }
  }
  Tcl_ResetResult(interp);
  return TCL_OK;
}

static int
  Handle_spawn_kill(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  SpawnData *spawn_data = 0;
  int argnum = 1;

  {
    int code = ParseOptions(interp,argc,argv,&argnum,&spawn_data);

    if (code!=TCL_OK)
      return code;
  }
  --argnum;
  {
    int code = GetNextArg(interp,argc,argv,&argnum,"signal name");
    if (code!=TCL_OK)
      return code;
  }
  if (argnum+1!=argc) {
    Tcl_SetResult(interp,"Extra parameters specified",TCL_STATIC);
    return TCL_ERROR;
  }
  {
    int i =0;
    for (;signals[i].name;++i) {
      if (strcmp(signals[i].name,argv[argnum])==0) {
	fprintf(
	  stderr,
	  "killing pid %d with %d\n",spawn_data->child_pid,signals[i].signum
	);
	kill(spawn_data->child_pid,signals[i].signum);
	return TCL_OK;
      }
    }
  }
  Tcl_SetResult(interp,"Invalid signal name",TCL_STATIC);
  return TCL_ERROR;
}

static int
  Handle_spawn_stty(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  SpawnData *spawn_data = 0;
  int argnum = 1;
  int code = ParseOptions(interp,argc,argv,&argnum,&spawn_data);
  int size_changed = 0;

  if (code!=TCL_OK)
    return code;

  for (;argnum<argc;++argnum) {
    if (strcmp(argv[argnum],"rows")==0) {
      {
	int code = GetNextArg(interp,argc,argv,&argnum,"rows");
	if (code!=TCL_OK)
	  return code;
      }
      {
	int rows;
	int code = Tcl_GetInt(interp,argv[argnum],&rows);
	if (code!=TCL_OK)
	  return code;
	spawn_data->window_size.ws_row = rows;
	size_changed = 1;
      }
    }
    else if (strcmp(argv[argnum],"cols")==0) {
      {
	int code = GetNextArg(interp,argc,argv,&argnum,"cols");
	if (code!=TCL_OK)
	  return code;
      }
      {
	int cols;
	int code = Tcl_GetInt(interp,argv[argnum],&cols);
	if (code!=TCL_OK)
	  return code;
	spawn_data->window_size.ws_col = cols;
	size_changed = 1;
      }
    }
    else {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"Invalid argument '",argv[argnum],"'",NULL);
      return TCL_ERROR;
    }
  }
  if (size_changed) {
    ioctl(spawn_data->slave_fd,TIOCSWINSZ,&spawn_data->window_size);
  }
  return TCL_OK;
}

static int
  Handle_spawn_open(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  SpawnData *spawn_data = NewSpawnData(&spawn_cmd_handlers);

  spawn_data->interp = interp;
  OpenPty(spawn_data);
  Tcl_SetVar2(interp,"spawn_out","slave,name",spawn_data->slave_name,0);
  {
    Tcl_Obj *spawn_id_obj = Tcl_NewIntObj(SpawnId(spawn_data));
    Tcl_IncrRefCount(spawn_id_obj);
    SetObjVar(interp,"spawn_id",spawn_id_obj);
    Tcl_SetObjResult(interp,spawn_id_obj);
    Tcl_DecrRefCount(spawn_id_obj);
  }
  Tcl_CreateFileHandler(
    spawn_data->master_fd,TCL_READABLE|TCL_EXCEPTION,ReadMaster,spawn_data
  );
  return TCL_OK;
}

void 
  SpawnSetDefaultProcs(
    SpawnData *spawn_data,CommandHandlers *default_procs,void *client_data
  )
{
  spawn_data->default_procs = *default_procs;
  spawn_data->default_data = client_data;
}

int Spawn_Init(Tcl_Interp *interp)
{
  struct {
    char *name;
    int (*proc)(ClientData,Tcl_Interp *,int,char **);
  } commands[] = {
    {"spawn_open",  Handle_spawn_open},
    {"spawn_exec",  Handle_spawn_exec},
    {"spawn_delete",Handle_spawn_delete},
    {"spawn_close", Handle_spawn_close},
    {"spawn_stty",  Handle_spawn_stty},
    {"spawn_kill",  Handle_spawn_kill},
    {"spawn_send",    Handle_spawn_send},
    {"spawn_set_handlers",Handle_spawn_set_handlers},
    {NULL,NULL}
  };
  {
    ClientData client_data = 0;
    Tcl_CmdDeleteProc *delete_proc = 0;

    int i = 0;
    for (;commands[i].name;++i) {
      Tcl_CreateCommand(
	interp,commands[i].name,commands[i].proc,client_data,delete_proc
      );
    }
  }

  assert(!orig_child_signal_handler);
  orig_child_signal_handler = signal(SIGCHLD,SIGCHLDHandler);
  if (orig_child_signal_handler==SIG_ERR) {
    fprintf(stderr,"Error installing signal handler\n");
    perror("signal");
  }
  assert(orig_child_signal_handler!=SIGCHLDHandler);
  async_handler = Tcl_AsyncCreate(AsyncHandler,(ClientData)0);
  return TCL_OK;
}

void CallTermHandlerInt(SpawnData *spawn_data,TermOp_t op,int arg)
{
  TermEvalInt(spawn_data,op,arg);
}

void CallTermHandler(SpawnData *spawn_data,TermOp_t op)
{
  TermEval(spawn_data,op);
}
