#include "termtcl.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "timer.h"

#define TIMING_TEST 0

#if TIMING_TEST
#define BEGIN_TIMING(x) BeginTiming(x)
#define END_TIMING(x) EndTiming(x)
#else
#define BEGIN_TIMING(x) ;
#define END_TIMING(x) ;
#endif

static int next_term_id = 1;
static int term_array_size = 0;
static Term **term_array = 0;

#if TIMING_TEST
static Timer call_command_timer;
static Timer term_element_timer;
static Timer term_set_element_timer;
static Timer term_insert_timer;
static Timer call_insert_handlers_timer;
static Timer text_delete_timer;
static Timer text_compare_timer;
static Timer text_insert_timer;
static Timer text_command_timer;
static Timer text_get_timer;
static Timer termptr_timer;
#endif

Term *TermPtr(Tcl_Interp *interp,char *termname)
{
  Term *term = 0;

  BEGIN_TIMING(&termptr_timer);
  {
    int term_id;
    int error = 0;

    if (!error && strlen(termname)<=4) {
      error = 1;
    }
    if (!error && strncmp(termname,"term",4)!=0) {
      error = 1;
    }
    if (!error) {
      term_id = atoi(termname+4);
    }
    if (!error && term_id<0) {
      error = 1;
    }
    if (!error && term_id>=term_array_size) {
      error = 1;
    }
    if (!error && !term_array[term_id]) {
      error = 1;
    }
    if (!error) {
      assert(term_array[term_id]->id==term_id);

      term = term_array[term_id];
    }
  }
  END_TIMING(&termptr_timer);

  if (!term) {
    Tcl_SetResult(interp,"Invalid term name",TCL_STATIC);
  }

  return term;
}

static void SetString(char **str_ptr,const char *value)
{
  if (*str_ptr) {
    free(*str_ptr);
  }
  if (value) {
    *str_ptr = strdup(value);
  }
  else {
    *str_ptr = 0;
  }
}

static int IntValue(Tcl_Interp *interp,char *str)
{
  int value;

  if (Tcl_GetInt(interp,str,&value)!=TCL_OK) {
    assert(0);
  }
  return value;
}

#if 0
static void 
  CallCommand(Tcl_Interp *interp,int argc,char **argv)
{
  Tcl_CmdInfo cmdinfo;

  BEGIN_TIMING(&call_command_timer);
  BEGIN_TIMING(&get_command_info_timer);
  if (!Tcl_GetCommandInfo(interp,argv[0],&cmdinfo)) {
    assert(0);
  }
  END_TIMING(&get_command_info_timer);
  Tcl_ResetResult(interp);
  {
    int result = (*cmdinfo.proc)(cmdinfo.clientData,interp,argc,argv);

    assert(result==TCL_OK);
    END_TIMING(&call_command_timer);
  }
}
#endif

static void TextCommand(Term *term,int argc,char **argv)
{
  BEGIN_TIMING(&text_command_timer);
  {
    Tcl_Interp *interp = term->interp;

    Tcl_ResetResult(interp);
    {
      int result =
	(*term->widget_proc)(term->widget_clientdata,interp,argc,argv);
      assert(result==TCL_OK);
    }
  }
  END_TIMING(&text_command_timer);
}

static char *TermElement(Tcl_Interp *interp,char *termname,char *varname)
{
  BEGIN_TIMING(&term_element_timer);
  {
    char *result = Tcl_GetVar2(interp,termname,varname,TCL_GLOBAL_ONLY);

    assert(result);
    END_TIMING(&term_element_timer);
    return result;
  }
}

static void 
  TermSetElement(
    Tcl_Interp *interp,
    const char *termname,
    const char *varname,
    const char *value
  )
{
  BEGIN_TIMING(&term_set_element_timer);
  if (!Tcl_SetVar2(
        interp,
	(char *)termname,
	(char *)varname,
	(char *)value,
	TCL_GLOBAL_ONLY
      )) {
    assert(0);
  }
  END_TIMING(&term_set_element_timer);
}

static void
  TermSetIntElement(Tcl_Interp *interp,char *termname,char *varname,int value)
{
  char strvalue[8];
  sprintf(strvalue,"%d",value);
  assert(strlen(strvalue)<sizeof(strvalue));
  TermSetElement(interp,termname,varname,strvalue);
}

static int TermIntElement(Tcl_Interp *interp,char *termname,char *varname)
{
  return IntValue(interp,TermElement(interp,termname,varname));
}

static int IntResult(Tcl_Interp *interp)
{
  Tcl_Obj *result_obj = Tcl_GetObjResult(interp);
  int result_int;

  if (Tcl_GetIntFromObj(interp,result_obj,&result_int)!=TCL_OK) {
    assert(0);
  }
  return result_int;
}

static int TermIsActive(Term *term)
{
  Tcl_Interp *interp = term->interp;
  int result;
  char *widget = term->widget_name;

  assert(widget);
  {
    char *active_term_list = 
      Tcl_GetVar2(interp,"active_terms",widget,TCL_GLOBAL_ONLY);
    assert(active_term_list);
    {
      int list_size;
      char **list_elements;
      if (
	Tcl_SplitList(interp,active_term_list,&list_size,&list_elements)
	!= TCL_OK
      ) {
	assert(0);
      }
      {
	int i = 0;
	for (;i<list_size;++i) {
	  if (strcmp(list_elements[i],term->termname)==0) {
	    break;
	  }
	}
	result = (i<list_size);
	Tcl_Free((char *)list_elements);
      }
    }
  }
  return result;
}

static int
  Handle_term_alloc(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==1);
  if (term_array_size<=next_term_id) {
    int old_array_size = term_array_size;

    if (!term_array) {
      term_array_size = 256;
      term_array = (Term **)malloc(term_array_size*sizeof(Term *));
    }
    else {
      term_array_size *= 2;
      term_array = (Term **)realloc(term_array,term_array_size*sizeof(Term *));
    }
    {
      int i = old_array_size;
      for (;i<term_array_size;++i) {
	term_array[i] = 0;
      }
    }
  }
  {
    Term *term = (Term *)malloc(sizeof(Term));
    char termname[16];

    sprintf(termname,"term%d",next_term_id);
    term->termname = 0;
    term->id = next_term_id;
    term->widget_name = 0;
    term->charset = 0;
    term->tags = 0;
    term->fgcolor = 0;
    term->bgcolor = 0;
    term->font_tags = 0;
    term->insert_mark = 0;
    term->font_styles[0] = USASCII_STYLE;
    term->font_styles[1] = USASCII_STYLE;
    term->insert_mode = 0;
    term->inverse = 0;
    term->cur_row = 0;
    term->cur_col = 0;
    term->row_offset = 0;
    term->scroll_top = 0;
    term->scroll_bottom = 0;
    term->screen_height = 0;
    term->screen_width = 0;
    term->update_pending = 0;
    term->interp = interp;
    SetString(&term->termname,termname);
    term_array[next_term_id] = term;
    strcpy(interp->result,term->termname);
  }
  ++next_term_id;
  return TCL_OK;
}

static const char *FocusedTerm(Tcl_Interp *interp,char *widget)
{
  return Tcl_GetVar2(interp,"focused_terms",widget,TCL_GLOBAL_ONLY);
}

static int
  Handle_term_dealloc(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    char *termname = argv[1];
    Term *term = TermPtr(interp,termname);

    if (!term) {
      return TCL_ERROR;
    }
    {
      int term_id = term->id;
      const char *focused_term = FocusedTerm(interp,term->widget_name);

      assert(!focused_term || strcmp(focused_term,termname)!=0);
      SetString(&term->widget_name,0);
      SetString(&term->termname,0);
      SetString(&term->tags,0);
      SetString(&term->fgcolor,0);
      SetString(&term->bgcolor,0);
      SetString(&term->font_tags,0);
      SetString(&term->insert_mark,0);
      term_array[term_id]->id = -1;
      free(term_array[term_id]);
      term_array[term_id] = 0;
    }
  }
  return TCL_OK;
}

static void TermUpdateTags(Term *term)
{
  char *fgcolor = term->fgcolor;
  char *bgcolor = term->bgcolor;
  char *font_tags = term->font_tags;

  if (term->inverse) {
    int new_tags_len =
      strlen(fgcolor)+sizeof("_background")+
      strlen(bgcolor)+sizeof("_foreground")+
      strlen(font_tags);
    char *new_tags = (char *)malloc(new_tags_len+1);

    sprintf(
      new_tags,"%s_background %s_foreground %s",fgcolor,bgcolor,font_tags
    );
    assert(strlen(new_tags)==new_tags_len);
    SetString(&term->tags,0);
    term->tags = new_tags;
  }
  else {
    int fglen = 0;
    int bglen = 0;

    if (strlen(fgcolor)>0) {
      fglen = strlen(fgcolor)+sizeof("_foreground");
    }
    if (strlen(bgcolor)>0) {
      bglen = strlen(bgcolor)+sizeof("_background");
    }
    {
      int new_tags_len = fglen+bglen+strlen(font_tags);
      char *new_tags = (char *)malloc(new_tags_len+1);
      int pos = 0;

      if (strlen(fgcolor)>0) {
	sprintf(new_tags+pos,"%s_foreground ",fgcolor);
	pos += fglen;
      }
      if (strlen(bgcolor)>0) {
	sprintf(new_tags+pos,"%s_background ",bgcolor);
	pos += bglen;
      }
      strcpy(new_tags+pos,font_tags);
      assert(strlen(new_tags)==new_tags_len);
      SetString(&term->tags,0);
      term->tags = new_tags;
    }
  }
}

static int
  Handle_term_update_tags(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermUpdateTags(term);
  }
  return TCL_OK;
}

void TermTcl_ResetTags(Term *term)
{
  SetString(&term->fgcolor,"");
  SetString(&term->bgcolor,"");
  term->inverse = 0;
  SetString(&term->font_tags,"");
  TermUpdateTags(term);
}

static int 
  Handle_term_reset_tags(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_ResetTags(term);
  }
  return TCL_OK;
}

void TermTcl_SetCharset(Term *term,int charset)
{
  term->charset = charset;
}

static int
  Handle_term_set_charset(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetCharset(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_SetInsertMode(Term *term,int value)
{
  term->insert_mode = value;
}

static int
  Handle_term_set_insertmode(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetInsertMode(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_SetInverse(Term *term,int value)
{
  term->inverse = value;
  TermUpdateTags(term);
}

static int
  Handle_term_set_inverse(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetInverse(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

    
static int
  Handle_term_is_active(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    if (TermIsActive(term)) {
      strcpy(interp->result,"1");
    }
    else {
      strcpy(interp->result,"0");
    }
  }
  return TCL_OK;
}

static int
  Handle_focused_term(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    char *widget = argv[1];
    const char *result = FocusedTerm(interp,widget);
    if (!result) {
      interp->result[0] = 0;
    }
    else {
      strcpy(interp->result,result);
    }
  }
  return TCL_OK;
}

static void TextSee(Term *term,char *index)
{
  char *args[3] = {0,"see",0};

  args[0] = term->widget_name;
  args[2] = index;

  TextCommand(term,3,args);
}

static int
  Handle_text_see(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TextSee(term,argv[2]);
  }
  return TCL_OK;
}

static void TermMarkRowCol(Term *term,char *index,int *rowptr,int *colptr)
{
  Tcl_Interp *interp = term->interp;

  {
    char *args[3] = {0,"index",0};

    args[0] = term->widget_name;
    args[2] = index;

    TextCommand(term,3,args);
  }
  if (sscanf(interp->result,"%d.%d",rowptr,colptr)!=2) {
    fprintf(stderr,"Invalid index\n");
    abort();
  }
}

static int
  Handle_term_mark_row_col(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==5);
  {
    char *termname = argv[1];
    char *index = argv[2];
    char *rowvar = argv[3];
    char *colvar = argv[4];
    int row,col;
    Term *term = TermPtr(interp,termname);

    if (!term) {
      return TCL_ERROR;
    }
    TermMarkRowCol(term,index,&row,&col);
    {
      char rowstr[16];
      char colstr[16];

      sprintf(rowstr,"%d",row);
      sprintf(colstr,"%d",col);
      Tcl_SetVar(interp,rowvar,rowstr,0);
      Tcl_SetVar(interp,colvar,colstr,0);
    }
  }
  return TCL_OK;
}

static int TermAllowedScroll(Term *term)
{
  Tcl_Interp *interp = term->interp;
  char *widget = term->widget_name;
  const char *focused_term = FocusedTerm(interp,widget);
  char *termname = term->termname;

  if ((focused_term && strcmp(focused_term,termname)==0) ||
      (!focused_term && !TermIsActive(term))) {
    return 1;

  } 
  else {
    return 0;
  }
}

static void TermSee(Term *term,char *index)
{
  if (TermAllowedScroll(term)) {
    TextSee(term,index);
  }
}

static char *TermInsertMark(Term *term)
{
  return term->insert_mark;
}

static void 
  TermSetInsertMark(Tcl_Interp *interp,Term *term,const char *value)
{
  SetString(&term->insert_mark,value);
}

static void TextInsert(Term *term,char *index,const char *text,char *tags)
{
  BEGIN_TIMING(&text_insert_timer);
  {
    char *args[5] = {0,"insert",0,0,0};

    args[0] = term->widget_name;
    args[2] = index;
    args[3] = (char *)text;
    args[4] = tags;
    
    TextCommand(term,5,args);
  }
  END_TIMING(&text_insert_timer);
}

static void TextGet(Term *term,char *index1,char *index2)
{
  BEGIN_TIMING(&text_get_timer);
  {
    char *args[4] = {0,"get",0,0};

    args[0] = term->widget_name;
    args[2] = index1;
    args[3] = index2;

    TextCommand(term,4,args);
  }
  END_TIMING(&text_get_timer);
}

static void TermPositionCursorRow(Term *term)
{
  char *widget = term->widget_name;
  char bottom_mark[32];
  char top_mark[32];
  int bottom_row,bottom_col;
  int top_row,top_col;
  int cur_row;
  char *termname = term->termname;

  assert(term->cur_row<term->screen_height);
  assert(term->cur_row>=0);
  assert(term->cur_col<term->screen_width);
  assert(term->cur_col>=0);

  sprintf(bottom_mark,"%sbottom",termname);
  sprintf(top_mark,"%stop",termname);
  TermMarkRowCol(term,bottom_mark,&bottom_row,&bottom_col);
  TermMarkRowCol(term,top_mark,&top_row,&top_col);
  cur_row = top_row + term->row_offset;
  cur_row += term->cur_row;
  {
    char bottom_prev_index[4+16+6+1+2+1+5+1+7];
    sprintf(bottom_prev_index,"%sbottom -1 lines lineend",termname);
    while (cur_row>=bottom_row) {
      TextInsert(term,bottom_prev_index,"\n","");
      TermSee(term,bottom_prev_index);
      ++bottom_row;
    }
  }
  {
    char *insert_mark = TermInsertMark(term);
    char row_index[20];

    sprintf(row_index,"%d.0",cur_row);
    {
      char *args[5] = {0,"mark","set",0,0};

      args[0] = widget;
      args[3] = insert_mark;
      args[4] = row_index;
      TextCommand(term,5,args);
    }
  }
}

#if 0
static void TermCheck(Term *term)
{
  int bottom_row;
  int bottom_col;
  int insert_row;
  int insert_col;
  char bottom_mark[32];
  sprintf(bottom_mark,"%sbottom",term->termname);

  TermMarkRowCol(term->interp,term,bottom_mark,&bottom_row,&bottom_col);
  TermMarkRowCol(
    term->interp,
    term,
    TermInsertMark(term),
    &insert_row,
    &insert_col
  );
  assert(insert_row<bottom_row);
}
#endif

static void TermPositionCursorCol(Term *term)
{
  Tcl_Interp *interp = term->interp;
  int cur_col = term->cur_col;
  char *term_widget = term->widget_name;
  char *term_insert = TermInsertMark(term);
  char linestart_index[64];
  sprintf(linestart_index,"%s linestart",term_insert);
  assert(strlen(linestart_index)<sizeof linestart_index);
  
  assert(cur_col<term->screen_width);
  if (cur_col==0) {
    {
      char *args[5] = {0,"mark","set",0,0};

      args[0] = term_widget;
      args[3] = term_insert;
      args[4] = linestart_index;
      TextCommand(term,5,args);
    }
  }
  else {
    int linelen = 0;
    char lineend_index[64];

    sprintf(lineend_index,"%s lineend",term_insert);
    assert(strlen(lineend_index)<sizeof lineend_index);
    TextGet(term,linestart_index,lineend_index);
    linelen = strlen(interp->result);

    if (linelen<cur_col) {
      int spacecount = cur_col-linelen;
      char *spaces = (char *)Tcl_Alloc(spacecount+1);
      sprintf(spaces,"%*s",spacecount,"");
      TextInsert(term,lineend_index,spaces,"");
      Tcl_Free(spaces);
    }
    {
      char new_insert_index[16+1+9+1+1+16+1+1];
      sprintf(new_insert_index,"%s linestart +%d c",term_insert,cur_col);
      assert(strlen(new_insert_index)<sizeof(new_insert_index));

      {
	char *args[5] = {0,"mark","set",0,0};

        args[0] = term_widget;
        args[3] = term_insert;
        args[4] = new_insert_index;
	TextCommand(term,5,args);
      }
    }
  }
}

static void TermPositionCursor(Term *term)
{
  TermPositionCursorRow(term);
  TermPositionCursorCol(term);
}

static int
  Handle_term_position_cursor(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermPositionCursor(term);
  }
  return TCL_OK;
}

void TermTcl_HomeCursor(Term *term)
{
  term->cur_row = 0;
  term->cur_col = 0;
  TermPositionCursor(term);
}

static int 
  Handle_term_home_cursor(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_HomeCursor(term);
  }
  return TCL_OK;
}

static int
  Handle_term_allowed_scroll(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    if (TermAllowedScroll(term)) {
      strcpy(interp->result,"1");
    }
    else {
      strcpy(interp->result,"0");
    }
  }
  return TCL_OK;
}

static int 
  Handle_term_see(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermSee(term,argv[2]);
  }
  return TCL_OK;
}

static int 
  Handle_term_position_cursor_row(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermPositionCursorRow(term);
  }
  return TCL_OK;
}

static int 
  Handle_term_position_cursor_col(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermPositionCursorCol(term);
  }
  return TCL_OK;
}

void TermTcl_SetFgColor(Term *term,int fgcolor)
{
  if (fgcolor==9) {
    SetString(&term->fgcolor,"");
  }
  else {
    char value[16];

    sprintf(value,"%d",fgcolor);
    SetString(&term->fgcolor,value);
  }
  TermUpdateTags(term);
}
  
void TermTcl_SetBgColor(Term *term,int bgcolor)
{
  if (bgcolor==9) {
    SetString(&term->bgcolor,"");
  }
  else {
    char value[16];

    sprintf(value,"%d",bgcolor);
    SetString(&term->bgcolor,value);
  }
  TermUpdateTags(term);
}
  
static int 
  Handle_term_set_fgcolor(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetFgColor(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

static int 
  Handle_term_set_bgcolor(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetBgColor(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

static char *ConvertDecSpecial(const char *s)
{
  char *ns = Tcl_Alloc(strlen(s)+1);
  int i = 0;

  for (;s[i];++i) {
    char c = s[i];

    if (c==0x5f) {
      c = 0x7f;
    }
    else if (c > 0x5f && c < 0x7f) {
      c -= 0x5f;
      if (c==10) {
	c = 0x7f;
      }
    }
    ns[i] = c;
  }
  ns[i] = '\0';
  return ns;
}

static int 
  Handle_convert_dec_special(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  if (argc<2) {
    Tcl_SetResult(interp,"not enough arguments",TCL_STATIC);
    return TCL_ERROR;
  }
  if (argc>2) {
    Tcl_SetResult(interp,"too many arguments",TCL_STATIC);
    return TCL_ERROR;
  }
  Tcl_SetResult(interp,ConvertDecSpecial(argv[1]),TCL_DYNAMIC);
  return TCL_OK;
}

void TermTcl_CR(Term *term)
{
  char *term_widget = term->widget_name;
  char *term_insert = TermInsertMark(term);
  char term_insert_linestart[32];

  sprintf(term_insert_linestart,"%s linestart",term_insert);
  assert(strlen(term_insert_linestart)<sizeof term_insert_linestart);
  term->cur_col = 0;
  {
    char *args[5] = {0,"mark","set",0,0};
    
    args[0] = term_widget;
    args[3] = term_insert;
    args[4] = term_insert_linestart;
    TextCommand(term,5,args);
  }
}

static int
  Handle_term_cr(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_CR(term);
  }
  return TCL_OK;
}

static void TermScroll(Term *term,int count)
{
  if (TermAllowedScroll(term)) {
    char count_str[16];
    char *term_widget = term->widget_name;

    sprintf(count_str,"%d",count);
    {
      char *args[5] = {0,"yview","scroll",0,"units"};
      
      args[0] = term_widget;
      args[3] = count_str;
      TextCommand(term,5,args);
    }
  }
}

static int
  Handle_term_scroll(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermScroll(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

static void TextDelete(Term *term,char *start,char *end)
{
  BEGIN_TIMING(&text_delete_timer);
  {
    char *args[4] = {0,"delete",0,0};

    args[0] = term->widget_name;
    args[2] = start;
    args[3] = end;
    TextCommand(term,4,args);
  }
  END_TIMING(&text_delete_timer);
}

static int 
  TextCompare(
    Tcl_Interp *interp,Term *term,char *index1,char *op,char *index2
  )
{
  int result;

  BEGIN_TIMING(&text_compare_timer);
  {
    char *args[5] = {0,"compare",0,0,0};
    
    args[0] = term->widget_name;
    args[2] = index1;
    args[3] = op;
    args[4] = index2;
    TextCommand(term,5,args);
  }
  result = IntResult(interp);
  END_TIMING(&text_compare_timer);

  return result;
}

static void TextMarkSet(Term *term,char *mark,char *index)
{
  char *args[5] = {0,"mark","set",0,0};
  args[0] = term->widget_name;
  args[3] = mark;
  args[4] = index;
  
  TextCommand(term,5,args);
}

static void TermOpenLine(Term *term)
{
  Tcl_Interp *interp = term->interp;
  char term_insert_lineend[32];
  char *term_widget = term->widget_name;
  char term_insert_plus_1_lines[64];
  char *term_insert = TermInsertMark(term);
  char *termname = term->termname;
  int term_scroll_top = term->scroll_top;
  int term_row_offset = term->row_offset;

  sprintf(term_insert_lineend,"%s lineend",term_insert);
  assert(strlen(term_insert_lineend)<sizeof term_insert_lineend);

  /* Put the insert mark at the end of the line so it will be carried */
  /* down one line by inserting the carriage return at the end */
  TextMarkSet(term,term_insert,term_insert_lineend);

  TextInsert(term,term_insert_lineend,"\n","");

  sprintf(term_insert_plus_1_lines,"%s +1 lines",term_insert);
  assert(strlen(term_insert_plus_1_lines)<sizeof term_insert_plus_1_lines);
  if (term_scroll_top!=0) {
    int top_row = term_scroll_top+term_row_offset;
    char top_index[64];
    char top_index2[64];
    char top_index_plus_1_lines[64];

    sprintf(top_index,"%stop +%d lines",termname,top_row);
    {
      char *args[3] = {0,"index",0};

      args[0] = term_widget;
      args[2] = top_index;
      TextCommand(term,3,args);
    }
    strcpy(top_index2,Tcl_GetStringResult(interp));
    assert(strlen(top_index2)<sizeof top_index2);
    sprintf(top_index_plus_1_lines,"%s +1 lines",top_index2);
    TextDelete(term,top_index2,top_index_plus_1_lines);
  }
  else {
    term->row_offset = term_row_offset+1;
    TermScroll(term,1);
  }
}

static int
  Handle_term_open_line(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermOpenLine(term);
  }
  return TCL_OK;
}

static void TermForceUpdate(Tcl_Interp *interp,char *term_widget)
{
  if (Tcl_GetVar2(interp,"term_update_pending",term_widget,TCL_GLOBAL_ONLY)) {
    Tcl_UnsetVar2(interp,"term_update_pending",term_widget,TCL_GLOBAL_ONLY);
  }
  while (Tcl_DoOneEvent(TCL_WINDOW_EVENTS|TCL_IDLE_EVENTS|TCL_DONT_WAIT)) {
  }
}

static int
  Handle_term_force_update(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  TermForceUpdate(interp,argv[1]);
  return TCL_OK;
}

typedef struct ForceUpdateData_s ForceUpdateData;

struct ForceUpdateData_s {
  Tcl_Interp *interp;
  char *term_widget;
};

static void TermForceUpdateProc(ClientData client_data)
{
  ForceUpdateData *data = (ForceUpdateData *)client_data;
  TermForceUpdate(data->interp,data->term_widget);
  Tcl_Free(data->term_widget);
  data->term_widget = 0;
  data->interp = 0;
  Tcl_Free((char *)data);
}

static void TermUpdate(Tcl_Interp *interp,char *term_widget)
{
  if (Tcl_GetVar2(interp,"term_update_pending",term_widget,TCL_GLOBAL_ONLY)) {
    return;
  }
  if (!Tcl_SetVar2(
	interp,"term_update_pending",term_widget,"1",TCL_GLOBAL_ONLY)) {
    assert(0);
  }
  {
    ForceUpdateData *data = 
      (ForceUpdateData *)Tcl_Alloc(sizeof(ForceUpdateData));
    data->interp = interp;
    data->term_widget = Tcl_Alloc(strlen(term_widget)+1);
    strcpy(data->term_widget,term_widget);
    Tcl_CreateTimerHandler(10,TermForceUpdateProc,(ClientData)data);
  }
}

static int
  Handle_term_update(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  TermUpdate(interp,argv[1]);
  return TCL_OK;
}

void TermTcl_MoveDown(Term *term)
{
  Tcl_Interp *interp = term->interp;
  int cur_row = term->cur_row;
  char *termname = term->termname;
  int scroll_bottom = term->scroll_bottom;
  char *term_widget = term->widget_name;

  if (cur_row == scroll_bottom) {
    TermOpenLine(term);
    TermUpdate(interp,term_widget);
  }
  else {
    int screen_height = term->screen_height;

    if (cur_row+1 == screen_height) {
      return;
    }
    assert(cur_row+1 < screen_height);
    ++cur_row;
    term->cur_row = cur_row;
    assert(term->cur_row<term->screen_height);
    {
      char *term_insert = TermInsertMark(term);
      char term_insert_lineend_plus_1c[64];
      char term_insert_lineend[32];
      char termname_bottom[32];

      sprintf(term_insert_lineend,"%s lineend",term_insert);
      assert(strlen(term_insert_lineend)<sizeof term_insert_lineend);
      sprintf(term_insert_lineend_plus_1c,"%s +1c",term_insert_lineend);
      assert(
	strlen(term_insert_lineend_plus_1c) <
	sizeof term_insert_lineend_plus_1c
      );
      sprintf(termname_bottom,"%sbottom",termname);
      assert(strlen(termname_bottom)<sizeof termname_bottom);

      if (TextCompare(
	    interp,
	    term,
	    term_insert_lineend_plus_1c,
	    ">=",
	    termname_bottom
	  )) {
	/* Put the insert mark at the end of the line so that inserting */
	/* the carriage return will take it down one line */
	TextMarkSet(term,term_insert,term_insert_lineend);
	TextInsert(term,term_insert_lineend,"\n","");
      }
      else {
	TextMarkSet(term,term_insert,term_insert_lineend_plus_1c);
      }
      TermSee(term,term_insert);
    }
  }
  TermPositionCursorCol(term);
}

static int
  Handle_term_move_down(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_MoveDown(term);
  }
  return TCL_OK;
}

static FontStyle_t TermFontStyle(Term *term,int charset)
{
  assert(charset>=0 && charset<=1);
  {
    FontStyle_t style = term->font_styles[charset];

    assert(style==DEC_SPECIAL_STYLE || style==USASCII_STYLE);
    return style;
  }
}

void TermTcl_SetFontStyle(Term *term,int charset,FontStyle_t style)
{
  assert(charset>=0 && charset<=1);
  assert(style>NO_STYLE && style<END_STYLE);
  term->font_styles[charset] = style;
}

static FontStyle_t FontStyleId(char *font_style_name)
{
  if (strcmp(font_style_name,"usascii")==0) {
    return USASCII_STYLE;
  }
  if (strcmp(font_style_name,"dec_special")==0) {
    return DEC_SPECIAL_STYLE;
  }
  assert(0);
}

static void 
  TermCallInsertHandlers(Tcl_Interp *interp,char *term_widget,char *termname)
{
  BEGIN_TIMING(&call_insert_handlers_timer);
  {
    char *insert_handler = 
      Tcl_GetVar2(interp,"term_insert_handlers",term_widget,TCL_GLOBAL_ONLY);

    if (insert_handler) {
      char *command = 0;
      int command_len = strlen(insert_handler)+1+strlen(termname);

      assert(insert_handler);
      command = Tcl_Alloc(command_len+1);
      sprintf(command,"%s %s",insert_handler,termname);
      assert(strlen(command)==command_len);

      if (Tcl_Eval(interp,command)!=TCL_OK) {
	assert(0);
      }
      Tcl_Free(command);
    }
  }
  END_TIMING(&call_insert_handlers_timer);
}

void TermTcl_Insert(Term *term,const char *orig_s,int slen)
{
  Tcl_Interp *interp = term->interp;
  char *termname = term->termname;
  char *new_s = 0;
  const char *s = 0;
  int charset = term->charset;
  char *term_widget = term->widget_name;
  char *term_insert = TermInsertMark(term);
  char term_insert_lineend[32];
  int screen_width = term->screen_width;
  int cur_col = term->cur_col;
  char *term_tags = term->tags;

  sprintf(term_insert_lineend,"%s lineend",term_insert);

  if (TermFontStyle(term,charset)==DEC_SPECIAL_STYLE) {
    new_s = ConvertDecSpecial(orig_s);
    s = new_s;
  }
  else {
    s = orig_s;
  }
  if (term->insert_mode) {
    if (cur_col+slen >= screen_width) {
      TextDelete(term,term_insert,term_insert_lineend);
    }
    else {
      if (TextCompare(interp,term,term_insert,"<",term_insert_lineend)) {
	TextGet(term,term_insert,term_insert_lineend);
	{
	  int linelen = strlen(Tcl_GetStringResult(interp));
	  if (slen+linelen>screen_width) {
	    int chop = slen+linelen-screen_width;
	    char term_insert_lineend_minus_chop_c[64];

	    sprintf(
	      term_insert_lineend_minus_chop_c,
	      "%s lineend -%d c",term_insert,chop
	    );
	    TextDelete(
	      term,
	      term_insert_lineend_minus_chop_c,
	      term_insert_lineend
	    );
	  }
	}
      }
    }
  }
  else {
    assert(cur_col<screen_width);
    while (cur_col+slen >= screen_width) {
      int inside = screen_width-cur_col;

      assert(inside>0);
      TextDelete(term,term_insert,term_insert_lineend);
      {
	char *s_inside = Tcl_Alloc(inside+1);

	memcpy(s_inside,s,inside);
	s_inside[inside] = '\0';
	TextInsert(term,term_insert,s_inside,term_tags);
	Tcl_Free(s_inside);
      }
      TermTcl_CR(term);
      cur_col = 0;
      TermTcl_MoveDown(term);
      {
	char term_insert_minus_1c[64];
	sprintf(term_insert_minus_1c,"%s -1c",term_insert);
	{
	  char *args[6] = {0,"tag","add","wrap",0,0};
	  
          args[0] = term_widget;
          args[4] = term_insert_minus_1c;
          args[5] = term_insert;
	  TextCommand(term,6,args);
	}
      }
      s += inside;
      slen -= inside;
    }
    if (slen>0 &&
        TextCompare(interp,term,term_insert,"<",term_insert_lineend)) {
      TextGet(term,term_insert,term_insert_lineend);
      {
	int linelen = strlen(Tcl_GetStringResult(interp));
	
	if (slen>=linelen) {
	  TextDelete(term,term_insert,term_insert_lineend);
	}
	else {
	  char term_insert_plus_slen_c[64];

	  sprintf(term_insert_plus_slen_c,"%s +%d c",term_insert,slen);
	  TextDelete(term,term_insert,term_insert_plus_slen_c);
	}
      }
    }
  }

  if (slen>0) {
    char *stmp = (char *)malloc(slen+1);

    memcpy(stmp,s,slen);
    stmp[slen] = 0;
    TextInsert(term,term_insert,stmp,term_tags);
    free(stmp);
  }

  assert(cur_col==term->cur_col);

  term->cur_col = cur_col+slen;

  TermCallInsertHandlers(interp,term_widget,termname);
  if (new_s) {
    Tcl_Free(new_s);
  }
  assert(term->cur_col<term->screen_width);
}

static int
  Handle_term_insert(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  BEGIN_TIMING(&term_insert_timer);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_Insert(term,argv[2],strlen(argv[2]));
  }
  END_TIMING(&term_insert_timer);
  return TCL_OK;
}

void TermTcl_SetFontTag(Term *term,char *tagname,int value)
{
  Tcl_Interp *interp = term->interp;
  char *font_tags = term->font_tags;

  fprintf(stderr,"TermSetFontTag: tagname='%s', value=%d\n",tagname,value);
  {
    int list_size;
    char **list_elements;
    char *new_font_tags = 0;

    if (
      Tcl_SplitList(interp,font_tags,&list_size,&list_elements)!=TCL_OK
    ) {
      assert(0);
    }

    {
      int i = 0;
      for (;i<list_size;++i) {
	if (strcmp(list_elements[i],tagname)==0) {
	  break;
	}
      }
      if (i<list_size) {
	if (!value) {
	  list_elements[i] = list_elements[list_size-1];
	  --list_size;
	  new_font_tags = Tcl_Merge(list_size,list_elements);
	}
      }
      else {
	if (value) {
	  ++list_size;
	  {
	    char **new_list_elements = 
	      (char **)Tcl_Alloc((list_size+1)*sizeof(char *));

	    {
	      int j = 0;
	      for (;j<list_size;++j) {
		new_list_elements[j] = list_elements[j];
	      }
	      new_list_elements[j] = tagname;
	    }
	    new_font_tags = Tcl_Merge(list_size+1,new_list_elements);
	    Tcl_Free((char *)new_list_elements);
	  }
	}
      }
    }
    if (new_font_tags) {
      SetString(&term->font_tags,new_font_tags);
      Tcl_Free(new_font_tags);
    }
    Tcl_Free((char *)list_elements);
  }
  TermUpdateTags(term);
}

static int
  Handle_term_set_font_tag(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==4);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetFontTag(term,argv[2],IntValue(interp,argv[3]));
  }
  return TCL_OK;
}

static int
  Handle_term_set_font_style(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==4);
  {
    int charset = IntValue(interp,argv[2]);
    FontStyle_t font_style = FontStyleId(argv[3]);
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetFontStyle(term,charset,font_style);
  }
  return TCL_OK;
}

static void TermSeeBottom(Term *term)
{
  char index[64];

  sprintf(index,"%sbottom -1 lines lineend",term->termname);
  assert(strlen(index)<sizeof index);
  TermSee(term,index);
}

void TermTcl_MoveUp(Term *term)
{
  Tcl_Interp *interp = term->interp;
  char *termname = term->termname;
  int cur_row = term->cur_row;
  int scroll_top = term->scroll_top;

  if (cur_row>scroll_top) {
    term->cur_row = cur_row-1;
    TermPositionCursor(term);
  }
  else {
    int scroll_bottom = term->scroll_bottom;
    int row_offset = term->row_offset;
    char *term_widget = term->widget_name;
    char bottom_index1[32];
    char bottom_index2[32];
    char bottom_index_plus_1_lines[32];

    sprintf(bottom_index1,"%stop +%d lines",termname,scroll_bottom+row_offset);
    assert(strlen(bottom_index1)<sizeof bottom_index1);
    {
      char *args[3] = {0,"index",0};

      args[0] = term_widget;
      args[2] = bottom_index1;
      TextCommand(term,3,args);
    }
    strcpy(bottom_index2,Tcl_GetStringResult(interp));
    assert(strlen(bottom_index2)<sizeof bottom_index2);
    sprintf(bottom_index_plus_1_lines,"%s +1 lines",bottom_index2);
    TextDelete(term,bottom_index2,bottom_index_plus_1_lines);
    {
      char index[32];

      sprintf(index,"%stop +%d lines",termname,row_offset+scroll_top);
      assert(strlen(index)<sizeof(index));
      TextInsert(term,index,"\n","");
    }
    TermPositionCursor(term);
    TermSeeBottom(term);
    TermUpdate(interp,term_widget);
  }
}

static int
  Handle_term_move_up(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_MoveUp(term);
  }
  return TCL_OK;
}

void TermTcl_InsertLine(Term *term,int count)
{
  char *termname = term->termname;
  int scroll_bottom = term->scroll_bottom;
  int row_offset = term->row_offset;
  int cur_row = term->cur_row;
  char bottom_index[64];
  char top_index[64];
  char bottom_index_plus_1_lines[64];

  sprintf(bottom_index,"%stop +%d lines",termname,scroll_bottom+row_offset);
  assert(strlen(bottom_index)<sizeof bottom_index);
  sprintf(top_index,"%stop +%d lines",termname,cur_row+row_offset);
  assert(strlen(top_index)<sizeof top_index);
  sprintf(
    bottom_index_plus_1_lines,
    "%stop +%d lines",
    termname,scroll_bottom+row_offset+1
  );
  assert(strlen(bottom_index_plus_1_lines)<sizeof bottom_index_plus_1_lines);

  while (count>0) {
    TextDelete(term,bottom_index,bottom_index_plus_1_lines);
    TextInsert(term,top_index,"\n","");
    TermSeeBottom(term);
    --count;
  }
}

static int
  Handle_term_insert_line(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_InsertLine(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_DeleteLine(Term *term,int count)
{
  char *termname = term->termname;
  int scroll_bottom = term->scroll_bottom;
  int row_offset = term->row_offset;
  int cur_row = term->cur_row;
  char bottom_index[64];
  char row_index[64];
  char row_index_plus_1_lines[64];

  sprintf(bottom_index,"%stop +%d lines",termname,scroll_bottom+row_offset);
  assert(strlen(bottom_index)<sizeof bottom_index);
  sprintf(row_index,"%stop +%d lines",termname,cur_row+row_offset);
  assert(strlen(row_index)<sizeof row_index);
  sprintf(
    row_index_plus_1_lines,
    "%stop +%d lines",
    termname,cur_row+row_offset+1
  );
  assert(strlen(row_index_plus_1_lines)<sizeof row_index_plus_1_lines);

  while (count>0) {
    TextDelete(term,row_index,row_index_plus_1_lines);
    TextInsert(term,bottom_index,"\n","");
    TermSeeBottom(term);
    --count;
  }
}

static int
  Handle_term_delete_line(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_DeleteLine(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_CursorUp(Term *term,int count)
{
  int cur_row = term->cur_row;
  int scroll_top = term->scroll_top;

  while (count>0) {
    if (cur_row==0 || cur_row==scroll_top) {
      break;
    }
    --cur_row;
    --count;
  }
  term->cur_row = cur_row;
  TermPositionCursor(term);
}

static int 
  Handle_term_cursor_up(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_CursorUp(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_CursorDown(Term *term,int count)
{
  int cur_row = term->cur_row;
  int screen_height = term->screen_height;
  int scroll_bottom = term->scroll_bottom;

  while (count>0) {
    if (cur_row+1==screen_height || cur_row==scroll_bottom) {
      break;
    }
    ++cur_row;
    --count;
  }
  term->cur_row = cur_row;
  TermPositionCursor(term);
}

static int 
  Handle_term_cursor_down(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_CursorDown(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_DeleteChar(Term *term,int count)
{
  Tcl_Interp *interp = term->interp;
  char *term_insert = TermInsertMark(term);
  char term_insert_lineend[32];

  sprintf(term_insert_lineend,"%s lineend",term_insert);
  TextGet(term,term_insert,term_insert_lineend);
  {
    int linelen = strlen(Tcl_GetStringResult(interp));

    if (linelen>=count) {
      char term_insert_plus_count_chars[32];

      sprintf(term_insert_plus_count_chars,"%s +%d c",term_insert,count);
      TextDelete(term,term_insert,term_insert_plus_count_chars);
    }
    else {
      TextDelete(term,term_insert,term_insert_lineend);
    }
  }
}
  
static int 
  Handle_term_delete_char(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_DeleteChar(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

static void TermShowUnused(Term *term)
{
  int cur_col = term->cur_col;
  int cur_row = term->cur_row;
  int screen_height = term->screen_height;

  term->cur_col = 0;
  term->cur_row = screen_height-1;
  TermPositionCursor(term);
  term->cur_col = cur_col;
  term->cur_row = cur_row;
  TermPositionCursor(term);
}

static int 
  Handle_term_show_unused(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermShowUnused(term);
  }
  return TCL_OK;
}

static void 
  TermClearScreenRange(
    Tcl_Interp *interp,Term *term,char *top_index,char *bottom_index
  )
{
  char *term_widget = term->widget_name;
  char index[16];

  {
    char *args[3] = {0,"index",0};

    args[0] = term_widget;
    args[2] = top_index;
    TextCommand(term,3,args);
  }
  strcpy(index,Tcl_GetStringResult(interp));
  for (;;) {
    if (!TextCompare(interp,term,index,"<",bottom_index))
      break;
    {
      char index_lineend[32];

      sprintf(index_lineend,"%s lineend",index);
      TextDelete(term,index,index_lineend);
    }
    {
      char index_plus_1_lines[32];

      sprintf(index_plus_1_lines,"%s +1 lines",index);
      {
	char *args[3] = {0,"index",0};
	
        args[0] = term_widget;
        args[2] = index_plus_1_lines;
	TextCommand(term,3,args);
      }
    }
    strcpy(index,Tcl_GetStringResult(interp));
  }
}

static int 
  Handle_term_clear_screen_range(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==4);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermClearScreenRange(interp,term,argv[2],argv[3]);
  }
  return TCL_OK;
}

void TermTcl_ClearBelow(Term *term)
{
  Tcl_Interp *interp = term->interp;
  char *term_widget = term->widget_name;
  char *term_insert = TermInsertMark(term);
  char bottom_minus_1_chars[32];
  char *termname = term->termname;

  sprintf(bottom_minus_1_chars,"%sbottom -1 c",termname);
  TermShowUnused(term);
  TermForceUpdate(interp,term_widget);
  TermClearScreenRange(interp,term,term_insert,bottom_minus_1_chars);
  TermPositionCursor(term);
}

static int 
  Handle_term_clear_below(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_ClearBelow(term);
  }
  return TCL_OK;
}

void TermTcl_ClearToRight(Term *term)
{
  char *term_insert = TermInsertMark(term);
  char term_insert_lineend[32];

  sprintf(term_insert_lineend,"%s lineend",term_insert);
  TextDelete(term,term_insert,term_insert_lineend);
}

static int 
  Handle_term_clear_to_right(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_ClearToRight(term);
  }
  return TCL_OK;
}

void TermTcl_ClearScreen(Term *term)
{
  Tcl_Interp *interp = term->interp;
  char *term_widget = term->widget_name;
  char top[32];
  char bottom[32];
  char *termname = term->termname;
  int row_offset = term->row_offset;

  sprintf(top,"%stop +%d lines",termname,row_offset);
  sprintf(bottom,"%sbottom",termname);

  TermForceUpdate(interp,term_widget);
  TermShowUnused(term);
  TermClearScreenRange(interp,term,top,bottom);
  TermPositionCursor(term);
}

static int 
  Handle_term_clear_screen(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_ClearScreen(term);
  }
  return TCL_OK;
}

void TermTcl_SaveCursor(Term *term)
{
  Tcl_Interp *interp = term->interp;
  int cur_row = term->cur_row;
  int cur_col = term->cur_col;
  char *termname = term->termname;

  TermSetIntElement(interp,termname,"saved_row",cur_row);
  TermSetIntElement(interp,termname,"saved_col",cur_col);
}

static int
  Handle_term_save_cursor(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SaveCursor(term);
  }
  return TCL_OK;
}

void TermTcl_RestoreCursor(Term *term)
{
  Tcl_Interp *interp = term->interp;
  char *termname = term->termname;
  int saved_row = TermIntElement(interp,termname,"saved_row");
  int saved_col = TermIntElement(interp,termname,"saved_col");

  term->cur_row = saved_row;
  term->cur_col = saved_col;
  TermPositionCursor(term);
}

static int
  Handle_term_restore_cursor(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_RestoreCursor(term);
  }
  return TCL_OK;
}

static void TermSimplePositionCursor(Term *term)
{
  char *termname = term->termname;
  char *term_widget = term->widget_name;
  char *term_insert = TermInsertMark(term);
  char index[64];
  int cur_row = term->cur_row;
  int cur_col = term->cur_col;
  int row_offset = term->row_offset;

  sprintf(index,"%stop +%d lines +%d c",termname,cur_row+row_offset,cur_col);
  {
    char *args[5] = {0,"mark","set",0,0};
    
    args[0] = term_widget;
    args[3] = term_insert;
    args[4] = index;
    TextCommand(term,5,args);
  }
}

static int
  Handle_term_simple_position_cursor(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermSimplePositionCursor(term);
  }
  return TCL_OK;
}

void TermTcl_Backspace(Term *term)
{
  int cur_col = term->cur_col;

  if (cur_col>0) {
    --cur_col;
    term->cur_col = cur_col;
    TermSimplePositionCursor(term);
  }
}

static int
  Handle_term_backspace(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_Backspace(term);
  }
  return TCL_OK;
}

void TermTcl_Tab(Term *term)
{
  int cur_col = term->cur_col;
  int screen_width = term->screen_width;

  cur_col += 8-(cur_col%8);
  while (cur_col>=screen_width) {
    ++term->cur_row;
    cur_col -= screen_width;
  }
  term->cur_col = cur_col;
  TermPositionCursor(term);
}

static int
  Handle_term_tab(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_Tab(term);
  }
  return TCL_OK;
}

void TermTcl_Unexpected(Term *term,int value)
{
  char bracketed_str[16];
  sprintf(bracketed_str,"<%d>",value);
  assert(strlen(bracketed_str)<sizeof bracketed_str);

  TermTcl_Insert(term,bracketed_str,strlen(bracketed_str));
  TermUpdate(term->interp,term->widget_name);
}

static int
  Handle_term_unexpected(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_Unexpected(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_SetRow(Term *term,int value)
{
  term->cur_row = value-1;
  TermPositionCursor(term);
}

static int
  Handle_term_set_row(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetRow(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_SetCol(Term *term,int value)
{
  term->cur_col = value-1;
  assert(term->cur_col>=0 && term->cur_col<term->screen_width);
  TermPositionCursorCol(term);
}

void TermTcl_InsertChar(Term *term,int count)
{
  Tcl_Interp *interp = term->interp;
  char *term_insert = TermInsertMark(term);
  char term_insert_lineend[32];
  char *term_tags = term->tags;

  sprintf(term_insert_lineend,"%s lineend",term_insert);
  TextGet(term,term_insert,term_insert_lineend);
  {
    int linelen = strlen(Tcl_GetStringResult(interp));
    int cur_col = term->cur_col;
    int screen_width = term->screen_width;
    int extra_chars = linelen+count+cur_col-screen_width;

    if (extra_chars>0) {
      char term_insert_lineend[32];

      sprintf(term_insert_lineend,"%s lineend",term_insert);
      if (extra_chars<linelen) {
	char start_index[32];

	sprintf(start_index,"%s lineend -%d c",term_insert,extra_chars);
	TextDelete(term,start_index,term_insert_lineend);
      }
      else {
	TextDelete(term,term_insert,term_insert_lineend);
      }
    }
    if (count+cur_col>screen_width) {
      count = screen_width-cur_col;
    }
    {
      char *chars = Tcl_Alloc(count+1);
      memset(chars,' ',count);
      chars[count] = 0;
      TextInsert(term,term_insert,chars,term_tags);
      Tcl_Free(chars);
    }
    TermPositionCursor(term);
  }
}

static int
  Handle_term_insert_char(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_InsertChar(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_CursorLeft(Term *term,int count)
{
  int cur_col = term->cur_col;

  cur_col -= count;
  if (cur_col<0) {
    cur_col = 0;
  }
  term->cur_col = cur_col;
  TermPositionCursor(term);
}

static int
  Handle_term_cursor_left(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_CursorLeft(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_CursorRight(Term *term,int count)
{
  int cur_col = term->cur_col;
  int screen_width = term->screen_width;

  cur_col += count;
  if (cur_col>=screen_width) {
    cur_col = screen_width-1;
  }
  term->cur_col = cur_col;
  TermPositionCursor(term);
}

static int
  Handle_term_cursor_right(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_CursorRight(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_SetCursor(Term *term,int row,int col)
{
  term->cur_row = row-1;
  term->cur_col = col-1;
  TermPositionCursor(term);
}

static int
  Handle_term_set_cursor(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==4);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetCursor(term,IntValue(interp,argv[2]),IntValue(interp,argv[3]));
  }
  return TCL_OK;
}

void TermTcl_SetScrollRegion(Term *term,int top,int bottom)
{
  term->scroll_top = top-1;
  term->scroll_bottom = bottom-1;
  term->cur_row = top-1;
  TermPositionCursor(term);
}

static int
  Handle_term_set_scroll_region(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==4);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetScrollRegion(
      term,IntValue(interp,argv[2]),IntValue(interp,argv[3])
    );
  }
  return TCL_OK;
}

void TermTcl_SetTab(Term *term,int value)
{
  /* Not implemented */
}

static int
  Handle_term_set_tab(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_SetTab(term,IntValue(interp,argv[2]));
  }
  return TCL_OK;
}

void TermTcl_ClearTabs(Term *term)
{
  /* Not implemented */
}

static int
  Handle_term_clear_tabs(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermTcl_ClearTabs(term);
  }
  return TCL_OK;
}

static void TermNewline(Term *term)
{
  term->cur_col = 0;
  TermSimplePositionCursor(term);
  TermTcl_MoveDown(term);
}

static int
  Handle_term_newline(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermNewline(term);
  }
  return TCL_OK;
}

static int
  Handle_term_widget(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    strcpy(interp->result,term->widget_name);
  }
  return TCL_OK;
}

static int
  Handle_term_set_widget(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Tcl_CmdInfo cmdinfo;
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }

    SetString(&term->widget_name,argv[2]);
    if (!Tcl_GetCommandInfo(interp,term->widget_name,&cmdinfo)) {
      assert(0);
    }
    term->widget_clientdata = cmdinfo.clientData;
    term->widget_proc = cmdinfo.proc;
  }
  return TCL_OK;
}

static int
  Handle_term_insert_mark(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    strcpy(interp->result,TermInsertMark(term));
  }
  return TCL_OK;
}

static int
  Handle_term_set_insert_mark(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    TermSetInsertMark(interp,term,argv[2]);
  }
  return TCL_OK;
}

static int
  Handle_term_set_insert_handler(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    char *widget = argv[1];
    char *handler = argv[2];

    if (strlen(handler)==0) {
      Tcl_UnsetVar2(interp,"term_insert_handlers",widget,TCL_GLOBAL_ONLY);
    }
    else {
      Tcl_SetVar2(interp,"term_insert_handlers",widget,handler,TCL_GLOBAL_ONLY);
    }
  }
  return TCL_OK;
}

static int
  Handle_term_cur_row(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    sprintf(interp->result,"%d",term->cur_row);
  }
  return TCL_OK;
}

static int
  Handle_term_cur_col(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    sprintf(interp->result,"%d",term->cur_col);
  }
  return TCL_OK;
}

static int
  Handle_term_set_cur_row(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    term->cur_row = IntValue(interp,argv[2]);
  }
  return TCL_OK;
}

static int
  Handle_term_set_cur_col(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    term->cur_col = IntValue(interp,argv[2]);
  }
  return TCL_OK;
}

static int
  Handle_term_row_offset(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    sprintf(interp->result,"%d",term->row_offset);
  }
  return TCL_OK;
}

static int
  Handle_term_set_row_offset(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    term->row_offset = IntValue(interp,argv[2]);
  }
  return TCL_OK;
}

static int
  Handle_term_set_scroll_top(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    term->scroll_top = IntValue(interp,argv[2]);
  }
  return TCL_OK;
}

static int
  Handle_term_set_scroll_bottom(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    term->scroll_bottom = IntValue(interp,argv[2]);
  }
  return TCL_OK;
}

static int
  Handle_term_scroll_top(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    sprintf(interp->result,"%d",term->scroll_top);
  }
  return TCL_OK;
}

static int
  Handle_term_scroll_bottom(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    sprintf(interp->result,"%d",term->scroll_bottom);
  }
  return TCL_OK;
}

static int
  Handle_term_set_screen_height(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    term->screen_height = IntValue(interp,argv[2]);
  }
  return TCL_OK;
}

static int
  Handle_term_screen_height(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    sprintf(interp->result,"%d",term->screen_height);
  }
  return TCL_OK;
}

static int
  Handle_term_screen_width(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==2);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    sprintf(interp->result,"%d",term->screen_width);
  }
  return TCL_OK;
}

static int
  Handle_term_set_screen_width(
    ClientData client_data,Tcl_Interp *interp,int argc,char **argv
  )
{
  assert(argc==3);
  {
    Term *term = TermPtr(interp,argv[1]);

    if (!term) {
      return TCL_ERROR;
    }
    term->screen_width = IntValue(interp,argv[2]);
  }
  return TCL_OK;
}

int TermTcl_Init(Tcl_Interp *interp)
{
  struct {
    char *name;
    int (*proc)(ClientData,Tcl_Interp *,int,char **);
  } commands[] = {
    {"term_alloc",              Handle_term_alloc},
    {"term_dealloc",            Handle_term_dealloc},
    {"term_home_cursor",        Handle_term_home_cursor},
    {"term_position_cursor",    Handle_term_position_cursor},
    {"term_mark_row_col",       Handle_term_mark_row_col},
    {"text_see",                Handle_text_see},
    {"focused_term",            Handle_focused_term},
    {"term_is_active",          Handle_term_is_active},
    {"term_allowed_scroll",     Handle_term_allowed_scroll},
    {"term_see",                Handle_term_see},
    {"term_position_cursor_row",Handle_term_position_cursor_row},
    {"term_position_cursor_col",Handle_term_position_cursor_col},
    {"term_update_tags",        Handle_term_update_tags},
    {"term_reset_tags",         Handle_term_reset_tags},
    {"term_set_fgcolor",        Handle_term_set_fgcolor},
    {"term_set_bgcolor",        Handle_term_set_bgcolor},
    {"convert_dec_special",     Handle_convert_dec_special},
    {"term_cr",                 Handle_term_cr},
    {"term_move_down",          Handle_term_move_down},
    {"term_scroll",             Handle_term_scroll},
    {"term_force_update",       Handle_term_force_update},
    {"term_update",             Handle_term_update},
    {"term_open_line",          Handle_term_open_line},
    {"term_insert",             Handle_term_insert},
    {"term_set_charset",        Handle_term_set_charset},
    {"term_set_insertmode",     Handle_term_set_insertmode},
    {"term_set_inverse",        Handle_term_set_inverse},
    {"term_set_font_tag",       Handle_term_set_font_tag},
    {"term_set_font_style",     Handle_term_set_font_style},
    {"term_move_up",            Handle_term_move_up},
    {"term_insert_line",        Handle_term_insert_line},
    {"term_delete_line",        Handle_term_delete_line},
    {"term_cursor_up",          Handle_term_cursor_up},
    {"term_cursor_down",        Handle_term_cursor_down},
    {"term_delete_char",        Handle_term_delete_char},
    {"term_show_unused",        Handle_term_show_unused},
    {"term_clear_screen_range", Handle_term_clear_screen_range},
    {"term_clear_below",        Handle_term_clear_below},
    {"term_clear_to_right",     Handle_term_clear_to_right},
    {"term_clear_screen",       Handle_term_clear_screen},
    {"term_save_cursor",        Handle_term_save_cursor},
    {"term_restore_cursor",     Handle_term_restore_cursor},
    {"term_backspace",          Handle_term_backspace},
    {"term_simple_position_cursor",Handle_term_simple_position_cursor},
    {"term_tab",                Handle_term_tab},
    {"term_unexpected",         Handle_term_unexpected},
    {"term_set_row",            Handle_term_set_row},
    {"term_insert_char",        Handle_term_insert_char},
    {"term_cursor_right",       Handle_term_cursor_right},
    {"term_cursor_left",        Handle_term_cursor_left},
    {"term_set_cursor",         Handle_term_set_cursor},
    {"term_set_scroll_region",  Handle_term_set_scroll_region},
    {"term_set_tab",            Handle_term_set_tab},
    {"term_clear_tabs",         Handle_term_clear_tabs},
    {"term_newline",            Handle_term_newline},
    {"term_widget",             Handle_term_widget},
    {"term_set_widget",         Handle_term_set_widget},
    {"term_insert_mark",        Handle_term_insert_mark},
    {"term_set_insert_mark",    Handle_term_set_insert_mark},
    {"term_set_insert_handler", Handle_term_set_insert_handler},
    {"term_cur_col",            Handle_term_cur_col},
    {"term_cur_row",            Handle_term_cur_row},
    {"term_set_cur_col",        Handle_term_set_cur_col},
    {"term_set_cur_row",        Handle_term_set_cur_row},
    {"term_row_offset",         Handle_term_row_offset},
    {"term_set_row_offset",     Handle_term_set_row_offset},
    {"term_set_scroll_top",     Handle_term_set_scroll_top},
    {"term_set_scroll_bottom",  Handle_term_set_scroll_bottom},
    {"term_scroll_top",         Handle_term_scroll_top},
    {"term_scroll_bottom",      Handle_term_scroll_bottom},
    {"term_set_screen_height",  Handle_term_set_screen_height},
    {"term_screen_height",      Handle_term_screen_height},
    {"term_screen_width",       Handle_term_screen_width},
    {"term_set_screen_width",   Handle_term_set_screen_width},
    {NULL,NULL}
  };
  {
    ClientData client_data = 0;
    Tcl_CmdDeleteProc *delete_proc = 0;

    int i = 0;
    for (;commands[i].name;++i) {
      Tcl_CreateCommand(
	interp,commands[i].name,commands[i].proc,client_data,delete_proc
      );
    }
  }
  return TCL_OK;
}

void TermTcl_PrintProfile(void)
{
#if TIMING_TEST
  PrintTimer(&call_command_timer,"CallCommand");
  PrintTimer(&term_element_timer,"TermElement");
  PrintTimer(&term_set_element_timer,"TermSetElement");
  PrintTimer(&term_insert_timer,"TermInsert");
  PrintTimer(&text_delete_timer,"TextDelete");
  PrintTimer(&text_compare_timer,"TextCompare");
  PrintTimer(&text_insert_timer,"TextInsert");
  PrintTimer(&text_command_timer,"TextCommand");
  PrintTimer(&text_get_timer,"TextGet");
  PrintTimer(&call_insert_handlers_timer,"TermCallInsertHandlers");
  PrintTimer(&termptr_timer,"TermPtr");
#endif
}
